/*******************************************************************************
 * Copyright (c) 2003 North Bears Team
 * All rights reserved.
 *
 * North Bears Team: Erick Antezana, Fabian Couche, Didier Croes, 
 * Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * 
 *******************************************************************************/

package graph.weighted;



import graph.weighted.weightedNodes.Graph;
import graph.weighted.weightedNodes.Node;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;

import north.graph.core.basic.BasicArc;
import north.graph.core.basic.BasicGraph;
import north.graph.core.basic.BasicNode;


/**
 * Class: GraphData Description: buffer between the biological data and the
 * graph structure
 *
 * @author Fabian Couche
 * @version $Revision: 1.19 $
 */
public class GraphData
  implements Serializable {

  /** Field substrateByReaction */
  private Hashtable _substrateByReaction;

  /** Field reactionBySubstrate */
  private Hashtable _reactionBySubstrate;

  /** Field productByReaction */
  private Hashtable _productByReaction;

  /** Field reactionByProduct */
  private Hashtable _reactionByProduct;

  /**  */
  private Hashtable _compoundWeight;
  
  protected HashMap modifiedWeights= null;

  /** GraphData() is the default constructor. */

  /**
   * GraphData Constructor
   */
  public GraphData () {

    _substrateByReaction = new Hashtable ();
    _reactionBySubstrate = new Hashtable ();
    _productByReaction = new Hashtable ();
    _reactionByProduct = new Hashtable ();
    _compoundWeight = new Hashtable ();
  }

  /**
   * Constructor.
   * @param fileName
  */
  public GraphData (String fileName) {
    this (fileName, false);
  }

  /**
   * Reads a saved GraphData using the parameter.
   *
   * @param fileName
   *            A String
   */
  public GraphData (String fileName, boolean ascii) {

    this ();

    // 
    assert fileName != null;

    // ascii or binary graph file are accepted
    if (ascii) {
      System.out.println ("before read ascii");
      this.readAsciiGraphData (fileName);

      this._substrateByReaction = invert (this._reactionBySubstrate);
      this._productByReaction = invert (this._reactionByProduct);

      System.out.println ("after read ascii");
    } else {
      System.out.println ("before read");

      GraphData temp = read (fileName);

      System.out.println ("after read");

      this._reactionBySubstrate = temp._reactionBySubstrate;
      this._substrateByReaction = invert (_reactionBySubstrate);

      System.out.println ("substrateByReaction find\t" + _substrateByReaction.size ());

      this._reactionByProduct = temp._reactionByProduct;
      this._productByReaction = invert (_reactionByProduct);

      System.out.println ("productByReaction find\t\t" + _productByReaction.size ());
    }
  }

  /**
   * Inverses a HashTable
   *
   * @param inTable
   *            A HashTable
   * @return the inversed HashTable
   */
  static protected Hashtable invert (Hashtable inTable) {

    // 
    assert inTable != null;

    // 
    Hashtable outTable = new Hashtable ();
    Object key, elem, x;
    ArrayList inVect, outVect;
    Enumeration input = inTable.keys ();

    while (input.hasMoreElements ()) {
      key = input.nextElement ();
      x = inTable.get (key);
      inVect = (ArrayList) x;

      Iterator elements = inVect.iterator ();

      while (elements.hasNext ()) {
        elem = elements.next ();

        if (outTable.containsKey (elem)) {
          outVect = (ArrayList) outTable.get (elem);

          outVect.add (key);
        } else {
          outVect = new ArrayList ();

          outVect.add (key);
          outTable.put (elem, outVect);
        }
      }
    }

    // 
    assert outTable != null;

    // 
    return (outTable);
  }

  /**
   * writeObject Method to serialize GraphData in binary
   *
   * @param out
   *            A Stream
   */
  protected void writeObject (ObjectOutputStream out) {

    try {
      out.writeObject (_reactionBySubstrate);
      out.writeObject (_reactionByProduct);
    } catch (Exception e) {
      System.out.println ("Exception during serialization: " + e);
      System.exit (0);
    }
  }

  /**
   * readObject Method to serialize GraphData in binary
   *
   * @param in
   *            A Stream
   * @throws ClassNotFoundException
   *             an exception.
   * @throws IOException
   *             an exception.
   */
  protected void readObject (java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException {

    _reactionBySubstrate = (Hashtable) in.readObject ();

    System.out.println ("reactionBySubstrate read\t" + _reactionBySubstrate.size ());

    _reactionByProduct = (Hashtable) in.readObject ();

    System.out.println ("reactionByProduct read\t\t" + _reactionByProduct.size ());
  }

  // public void setSubstrateByReaction(Map SR) {
  // substrateByReaction.putAll(SR);
  // }

  /**
   * Sets the table of reaction by substrate.
   *
   * @param rs
   *            the map to put.
   */
  public void setReactionBySubstrate (Map rs) {

    // 
    assert rs != null;

    // 
    _reactionBySubstrate.putAll (rs);
  }

  // public void setProductByReaction(Map PR) {
  // productByReaction.putAll(PR);
  // }

  /**
   * Sets the table of reaction by product.
   *
   * @param rp
   *            the map to put.
   */
  public void setReactionByProduct (Map rp) {

    // 
    assert rp != null;

    // 
    _reactionByProduct.putAll (rp);
  }

  /**
   * Constructs a new graph with the GraphData fields.
   *
   * @param defaultWeight
   *            the default weight.
   * @return A Grapth
   */
  public Graph getGraph (float defaultWeight) {

    Graph g = new Graph ();
    int nbArcs = 0;
    int nbNodes = 0;
    String react, comp;
    ArrayList vectInput, vectOutput;
    LinkedList listInput, listOutput = null;
    Enumeration inputKeys = _reactionBySubstrate.keys ();
    Hashtable reactionByProduct = new Hashtable ();

    reactionByProduct.putAll (_reactionByProduct);

    while (inputKeys.hasMoreElements ()) {
      comp = (String) inputKeys.nextElement ();
      vectInput = (ArrayList) _reactionBySubstrate.get (comp);

      if (reactionByProduct.containsKey (comp)) {
        vectOutput = (ArrayList) reactionByProduct.get (comp);

        reactionByProduct.remove (comp);
      }

      g.addCompound (comp, defaultWeight);

      nbNodes++;
    }

    Enumeration outputKeys = reactionByProduct.keys ();

    while (outputKeys.hasMoreElements ()) {
      comp = (String) outputKeys.nextElement ();
      vectOutput = (ArrayList) reactionByProduct.get (comp);

      g.addCompound (comp, defaultWeight);

      nbNodes++;
    }

    Hashtable productByReaction = new Hashtable ();

    productByReaction.putAll (_productByReaction);

    Enumeration substratesKeys = _substrateByReaction.keys ();

    while (substratesKeys.hasMoreElements ()) {
      react = (String) substratesKeys.nextElement ();

      // if (_productByReaction.containsKey (react)) {
      vectInput = (ArrayList) _substrateByReaction.get (react);
      vectOutput = (ArrayList) productByReaction.get (react);

      productByReaction.remove (react);

      listInput = new LinkedList (vectInput);

      if (vectOutput != null) {
        listOutput = new LinkedList (vectOutput);
        nbArcs += (listInput.size () + listOutput.size ());
      } else {
        nbArcs += listInput.size ();
      }

      g.addReaction (react, defaultWeight, listInput, listOutput);

      nbNodes += 2;

      // }
    }

    vectInput = null;

    Enumeration productsKeys = productByReaction.keys ();

    while (productsKeys.hasMoreElements ()) {
      react = (String) productsKeys.nextElement ();
      vectOutput = (ArrayList) productByReaction.get (react);
      listOutput = new LinkedList (vectOutput);
      nbArcs += listOutput.size ();

      g.addReaction (react, defaultWeight, null, listOutput);

      nbNodes += 2;
    }

    System.out.println ("Number of nodes: " + nbNodes);
    System.out.println ("Number of arcs: " + nbArcs);

    // 
    assert g != null;

    // 
    return g;
  }

  /**
   * Constructs a new graph with the GraphData fields
   *
   * @return A Graph
   */
  public Graph getGraph () {
    return this.getWeightedGraph (1);
  }

  /**
   *
   * @param reactionsWeight
   * @return
   */
  public Graph getWeightedGraph (float reactionsWeight) {

    Graph g = new Graph ();
    float w;
    int nbArcs = 0;
    int nbNodes = 0;
    String react, comp;
    ArrayList vectInput, vectOutput;
    LinkedList listInput, listOutput = null;
    Enumeration inputKeys = _reactionBySubstrate.keys ();
    Hashtable reactionByProduct = new Hashtable ();

    reactionByProduct.putAll (_reactionByProduct);

    while (inputKeys.hasMoreElements ()) {
      comp = (String) inputKeys.nextElement ();
      vectInput = (ArrayList) _reactionBySubstrate.get (comp);
      w = vectInput.size ();

      if (reactionByProduct.containsKey (comp)) {
        vectOutput = (ArrayList) reactionByProduct.get (comp);
        w += vectOutput.size ();

        reactionByProduct.remove (comp);
      }

      g.addCompound (comp, w);

      nbNodes++;
    }

    Enumeration outputKeys = reactionByProduct.keys ();

    while (outputKeys.hasMoreElements ()) {
      comp = (String) outputKeys.nextElement ();
      vectOutput = (ArrayList) reactionByProduct.get (comp);
      w = vectOutput.size ();

      g.addCompound (comp, w);

      nbNodes++;
    }

    Enumeration substratesKeys = _substrateByReaction.keys ();
    Hashtable productByReaction = new Hashtable ();

    productByReaction.putAll (_productByReaction);

    while (substratesKeys.hasMoreElements ()) {
      react = (String) substratesKeys.nextElement ();

      // if (_productByReaction.containsKey (react)) ?{
      vectInput = (ArrayList) _substrateByReaction.get (react);
      vectOutput = (ArrayList) _productByReaction.get (react);

      productByReaction.remove (react);

      listInput = new LinkedList (vectInput);

      if (vectOutput != null) {
        listOutput = new LinkedList (vectOutput);
      }

      nbArcs += (listInput.size () + listOutput.size ());

      g.addReaction (react, reactionsWeight, listInput, listOutput);

      nbNodes += 2;

      // }
    }

    vectInput = null;

    Enumeration productsKeys = productByReaction.keys ();

    while (productsKeys.hasMoreElements ()) {
      react = (String) productsKeys.nextElement ();
      vectOutput = (ArrayList) productByReaction.get (react);
      listOutput = new LinkedList (vectOutput);
      nbArcs += listOutput.size ();

      g.addReaction (react, reactionsWeight, null, listOutput);

      nbNodes += 2;
    }

    /*
     * Iterator reactionsItr = (new TreeSet (_substrateByReaction.keySet
     * ())).iterator (); Node currentReactionNode = null; // while
     * (substratesKeys.hasMoreElements ()) { // join all subreaction sharing
     * the same properties and marking while (reactionsItr.hasNext ()) {
     * react = (String) reactionsItr.next ();
     *
     * if (_productByReaction.containsKey (react)) { vectInput = (Vector)
     * _substrateByReaction.get (react); vectOutput = (Vector)
     * _productByReaction.get (react); listInput = new LinkedList
     * (vectInput); listOutput = new LinkedList (vectOutput); nbArcs +=
     * (listInput.size () + listOutput.size ());
     *
     * if ((currentReactionNode == null) || !react.startsWith
     * (currentReactionNode.getId ())) { int pos = react.indexOf ('_');
     *
     * if (pos > -1) { // if reaction are divided in subreactions
     * currentReactionNode = g.addReaction (react.substring (0, pos), 1,
     * listInput, listOutput); } else { // if reaction are not divided in
     * subreactions currentReactionNode = g.addReaction (react, 1,
     * listInput, listOutput); } } else {
     *
     * g.addReaction (currentReactionNode, listInput, listOutput); }
     *
     * nbNodes += 2; }
     */
    System.out.println ("Number of nodes: " + nbNodes);
    System.out.println ("Number of arcs: " + nbArcs);

    // 
    assert g != null;

    // 
    return g;
  }

  /**
   * return the connectivity of one compound
   *
   * @param compoundid
   *            the id of the compound
   * @return the connectivity of the compound
   */
  public float getCompoundConnectivity (String compoundid) {

    float w = -1;
    ArrayList vectInput = (ArrayList) _reactionBySubstrate.get (compoundid);

    if (vectInput != null) {
      w = vectInput.size ();
    }

    vectInput = (ArrayList) _reactionByProduct.get (compoundid);

    if (vectInput != null) {
      if (w == -1) {
        w = vectInput.size ();
      } else {
        w += vectInput.size ();
      }
    }

    return w;
  }

  /**
   * Displays all elements of hashtable.
   *
   * @param hTable
   *            a HashTable
   */
  public void displayHT (Hashtable hTable) {

    // 
    assert hTable != null;

    // 
    ArrayList vect;
    String vKey, elem;
    Enumeration enum = hTable.keys ();

    System.out.println ("Analyse keys:");

    while (enum.hasMoreElements ()) {
      vKey = (String) enum.nextElement ();

      System.out.println ("Key: " + vKey);

      vect = (ArrayList) hTable.get (vKey);

      for (int x = 0; x < vect.size (); ++x) {
        elem = (String) vect.get (x);

        System.out.print (elem + " ");
      }

      System.out.println ("");
    }

    System.out.println ("");
  }

  /**
   * Displays all informations of GraphData
   */
  public void display () {

    System.out.println ("substrateByReaction:");
    displayHT (_substrateByReaction);
    System.out.println ("");
    System.out.println ("reactionBySubstrate:");
    displayHT (_reactionBySubstrate);
    System.out.println ("");
    System.out.println ("productByReaction:");
    displayHT (_productByReaction);
    System.out.println ("");
    System.out.println ("reactionByProduct:");
    displayHT (_reactionByProduct);
    System.out.println ("");
  }

  /**
   * Saves a GraphData in a binary file.
   *
   * @param fileName
   *            the name of the file.
   */
  public void store (String fileName) {

    // 
    assert fileName != null;

    // 
    GraphData gData = this;

    try {
      FileOutputStream fos = new FileOutputStream (fileName);
      ObjectOutputStream oos = new ObjectOutputStream (fos);

      oos.writeObject (gData);
      oos.flush ();
      oos.close ();
    } catch (Exception e) {
      System.out.println ("Exception during serialization: " + e);
      System.exit (0);
    }
  }

  /**
   * Read a graphdata ascii file
   *
   * @param fileName
   */
  public void readAsciiGraphData (String fileName) {

    try {
      System.out.println ("reading graph : " + fileName);

      BufferedReader br = new BufferedReader (new FileReader (fileName));
      String s = br.readLine ();
      int linenb = 1;
      Hashtable reactionByCpd;

      for (int i = 0; i < 2; i++) {
        if (s != null) {
          reactionByCpd = new Hashtable ();

          String sCpd = null;
          ArrayList reactions = new ArrayList ();

          System.out.println (s);

          while ((s = br.readLine ()) != null && !s.equals ("--REACTION_BY_PRODUCT--")) {
            linenb++;

            StringTokenizer str = new StringTokenizer (s, "\t");
            String temp = str.nextToken ();

            if ((sCpd != null) && !sCpd.equals (temp)) {
              if (reactionByCpd.containsKey (sCpd)) {
                ((ArrayList) reactionByCpd.get (sCpd)).addAll (reactions);
              } else {
                reactionByCpd.put (sCpd, reactions);
              }

              reactions = new ArrayList ();
              sCpd = temp;
            } else {
              if (sCpd == null) {
                reactions = new ArrayList ();
                sCpd = temp;
              }
            }

            String reaction = str.nextToken ();

            reactions.add (reaction);
          }

          if (reactionByCpd.containsKey (sCpd)) {
            ((ArrayList) reactionByCpd.get (sCpd)).addAll (reactions);
          } else {
            reactionByCpd.put (sCpd, reactions);
          }

          if (i == 0) {
            this.setReactionBySubstrate (reactionByCpd);
          } else {
            this.setReactionByProduct (reactionByCpd);
          }
        } else {
          throw new IOException ("File is empty");
        }
      }

      System.out.println ("number of lines :" + linenb);
    } catch (IOException e) {
      System.out.println ("Exception during reading ascii file: " + e);
      System.exit (-1);
    }
  }

  /**
   * save graphdata in an tab delimited ascii file
   * @param fileName
   */
  public void saveAsciiGraph (String fileName) {

    try {
      int linenb = 1;
      BufferedWriter bw = new BufferedWriter (new FileWriter (fileName));
      Iterator itr = _reactionBySubstrate.entrySet ().iterator ();

      bw.write ("--REACTION_BY_SUBSTRATE--\n");

      while (itr.hasNext ()) {
        Map.Entry entry = (Map.Entry) itr.next ();
        String in = (String) entry.getKey ();
        Iterator reactionItr = ((ArrayList) entry.getValue ()).iterator ();

        while (reactionItr.hasNext ()) {
          bw.write (in + "\t" + reactionItr.next () + "\n");

          linenb++;
        }
      }

      itr = _reactionByProduct.entrySet ().iterator ();

      bw.write ("--REACTION_BY_PRODUCT--\n");

      linenb++;

      while (itr.hasNext ()) {
        Map.Entry entry = (Map.Entry) itr.next ();
        String in = (String) entry.getKey ();
        Iterator reactionItr = ((ArrayList) entry.getValue ()).iterator ();

        while (reactionItr.hasNext ()) {
          bw.write (in + "\t" + reactionItr.next () + "\n");

          linenb++;
        }
      }

      bw.flush ();
      bw.close ();
      System.out.println ("number of lines :" + linenb);
    } catch (IOException e) {
      System.out.println ("Exception during writing ascii file: " + e);
      System.exit (-1);
    }
  }

  /**
   * Reads a serialized GraphData in a file
   *
   * @param fileName
   *            A String
   * @return a GraphData
   */
  public GraphData read (String fileName) {

    // 
    assert fileName != null;

    // 
    GraphData back = new GraphData ();

    try {
      FileInputStream fis = new FileInputStream (fileName);
      ObjectInputStream ois = new ObjectInputStream (fis);

      back = (GraphData) ois.readObject ();

      ois.close ();
    } catch (Exception e) {
      System.out.println ("Exception during deserialization: " + e);
      System.exit (0);
    }

    // 
    assert back != null;

    // 
    return (back);
  }

  /**
   * Print the hashTables.
   */
  public void printTables () {

    System.out.println ("");
    System.out.println ("");
    System.out.println ("_reactionBySubstrate");

    Set keys = _reactionBySubstrate.keySet ();

    for (Iterator iter = keys.iterator (); iter.hasNext (); ) {
      String key = (String) iter.next ();

      System.out.println ("Key: " + key);
      System.out.print ("Values: ");

      ArrayList values = (ArrayList) _reactionBySubstrate.get (key);

      for (Iterator iterator = values.iterator (); iterator.hasNext (); ) {
        String value = (String) iterator.next ();

        System.out.print (value + " ");
      }

      System.out.println ("");
    }

    System.out.println ("");
    System.out.println ("");
    System.out.println ("_reactionByProduct");

    keys = _reactionByProduct.keySet ();

    for (Iterator iter = keys.iterator (); iter.hasNext (); ) {
      String key = (String) iter.next ();

      System.out.println ("Key: " + key);
      System.out.print ("Values: ");

      ArrayList values = (ArrayList) _reactionByProduct.get (key);

      for (Iterator iterator = values.iterator (); iterator.hasNext (); ) {
        String value = (String) iterator.next ();

        System.out.print (value + " ");
      }

      System.out.println ("");
    }

    System.out.println ("");
    System.out.println ("");
    System.out.println ("_substrateByReaction");

    keys = _substrateByReaction.keySet ();

    for (Iterator iter = keys.iterator (); iter.hasNext (); ) {
      String key = (String) iter.next ();

      System.out.println ("Key: " + key);
      System.out.print ("Values: ");

      ArrayList values = (ArrayList) _substrateByReaction.get (key);

      for (Iterator iterator = values.iterator (); iterator.hasNext (); ) {
        String value = (String) iterator.next ();

        System.out.print (value + " ");
      }

      System.out.println ("");
    }

    System.out.println ("");
    System.out.println ("");
    System.out.println ("_productByReaction");

    keys = _productByReaction.keySet ();

    for (Iterator iter = keys.iterator (); iter.hasNext (); ) {
      String key = (String) iter.next ();

      System.out.println ("Key: " + key);
      System.out.print ("Values: ");

      ArrayList values = (ArrayList) _productByReaction.get (key);

      for (Iterator iterator = values.iterator (); iterator.hasNext (); ) {
        String value = (String) iterator.next ();

        System.out.print (value + " ");
      }

      System.out.println ("");
    }

    System.out.println ("");
  }

  /**
   * Reads a invert reaction by substrate and reaction by product in a file
   */
  public void invertTables () {
    this._substrateByReaction = invert (_reactionBySubstrate);
    this._productByReaction = invert (_reactionByProduct);
  }

  /**
   * @return
   */
  public Hashtable getProductByReaction () {
    return _productByReaction;
  }

  /**
   * @return
   */
  public Hashtable getSubstrateByReaction () {
    return _substrateByReaction;
  }

  /**
   * @return Returns the reactionByProduct.
   */
  public Hashtable getReactionByProduct () {
    return _reactionByProduct;
  }

  /**
   * @return Returns the reactionBySubstrate.
   */
  public Hashtable getReactionBySubstrate () {
    return _reactionBySubstrate;
  }

  /**
   * @param substrateByReaction
   *            The substrateByReaction to set.
   */
  public void setSubstrateByReaction (Hashtable substrateByReaction) {
    _substrateByReaction = substrateByReaction;
  }

  /**
   * Build a basic graphe (reactions are not duplicated)
   * @return
   */
  public BasicGraph toBasicGraph () {

    BasicGraph bg = BasicGraph.newBasicGraph ();
    Iterator itr = this.getReactionByProduct ().entrySet ().iterator ();

    while (itr.hasNext ()) {
      Map.Entry entry = (Map.Entry) itr.next ();
      String cpdid = (String) entry.getKey ();
      int weight = (int) ((2000 / this.getCompoundConnectivity (cpdid))) + 1;
      ArrayList value = (ArrayList) entry.getValue ();
      BasicNode cpdn = null;

      if (bg.hasNode (cpdid)) {
        cpdn = bg.getNode (cpdid);
      } else {
        cpdn = BasicNode.newBasicNode (bg, cpdid, "", "");

        // cpdn =BasicNode.newBasicNode(cpdid);
        cpdn.putAttribute ("shape", "ellipse");
        bg.addNode (cpdn);
      }

      Iterator cpitr = value.iterator ();

      while (cpitr.hasNext ()) {
        String rid = (String) cpitr.next ();
        BasicNode rn = null;

        if (bg.hasNode (rid)) {
          rn = bg.getNode (rid);
        } else {
          rn = BasicNode.newBasicNode (bg, rid, "", "");

          rn.putAttribute ("shape", "box");
          bg.addNode (rn);
        }

        BasicArc ba = BasicArc.newBasicArc (rn, cpdn);

        ba.putAttribute ("weight", Integer.toString (weight));
        bg.addArc (ba);
      }
    }

    itr = this.getReactionBySubstrate ().entrySet ().iterator ();

    while (itr.hasNext ()) {
      Map.Entry entry = (Map.Entry) itr.next ();
      String cpdid = (String) entry.getKey ();
      int weight = (int) ((2000 / this.getCompoundConnectivity (cpdid))) + 1;
      ArrayList value = (ArrayList) entry.getValue ();
      BasicNode cpdn = null;

      if (bg.hasNode (cpdid)) {
        cpdn = bg.getNode (cpdid);
      } else {
        cpdn = BasicNode.newBasicNode (bg, cpdid, "", "");

        cpdn.putAttribute ("shape", "ellipse");
        bg.addNode (cpdn);
      }

      Iterator cpitr = value.iterator ();

      while (cpitr.hasNext ()) {
        String rid = (String) cpitr.next ();
        BasicNode rn = null;

        if (bg.hasNode (rid)) {
          rn = bg.getNode (rid);
        } else {
          rn = BasicNode.newBasicNode (bg, rid, "", "");

          rn.putAttribute ("shape", "box");
          bg.addNode (rn);
        }

        BasicArc ba = BasicArc.newBasicArc (cpdn, rn);

        ba.putAttribute ("weight", Integer.toString (weight));
        bg.addArc (ba);
      }
    }

    return bg;
  }

  /**
   * convert the graph in neato format
   * @return
   */
  public String toNeato () {

    String graph = "graph G { \n node [shape=box]; \n";
    TreeSet cpdSet = new TreeSet ();
    StringBuffer cpdnodes = new StringBuffer ("{node [shape=ellipse] ");
    StringBuffer edges = new StringBuffer ("");
    Iterator itr = this.getReactionBySubstrate ().entrySet ().iterator ();

    while (itr.hasNext ()) {
      Map.Entry entry = (Map.Entry) itr.next ();
      String cpdid = (String) entry.getKey ();

      cpdSet.add (cpdid);

      int weight = (int) ((2000 / this.getCompoundConnectivity (cpdid))) + 1;
      ArrayList value = (ArrayList) entry.getValue ();
      BasicNode cpdn = null;
      Iterator cpitr = value.iterator ();

      while (cpitr.hasNext ()) {
        String rid = (String) cpitr.next ();

        edges.append (cpdid + " -- " + rid + " [ weight=" + weight + " ];");
      }
    }

    itr = this.getReactionByProduct ().entrySet ().iterator ();

    while (itr.hasNext ()) {
      Map.Entry entry = (Map.Entry) itr.next ();
      String cpdid = (String) entry.getKey ();

      cpdSet.add (cpdid);

      int weight = (int) ((2000 / this.getCompoundConnectivity (cpdid))) + 1;
      ArrayList value = (ArrayList) entry.getValue ();
      BasicNode cpdn = null;
      Iterator cpitr = value.iterator ();

      while (cpitr.hasNext ()) {
        String rid = (String) cpitr.next ();

        edges.append (cpdid + " -- " + rid + " [ weight=" + weight + " ];");
      }
    }

    itr = cpdSet.iterator ();

    while (itr.hasNext ()) {
      cpdnodes.append (itr.next () + " ");
    }
    ;

    cpdnodes.append ("}\n");

    return graph + cpdnodes.toString () + edges.toString () + "}";
  }

  /**
   * convert the graphe in a neato light version
   * @return
   */
  public String toNeatoLight () {

    String graph = "graph G { \n";
    StringBuffer edges = new StringBuffer ("");
    Iterator itr = this.getReactionBySubstrate ().entrySet ().iterator ();

    while (itr.hasNext ()) {
      Map.Entry entry = (Map.Entry) itr.next ();
      String cpdid1 = (String) entry.getKey ();
      int weight1 = (int) ((2000 / this.getCompoundConnectivity (cpdid1))) + 1;
      ArrayList value = (ArrayList) entry.getValue ();
      BasicNode cpdn = null;
      Iterator cpitr = value.iterator ();

      while (cpitr.hasNext ()) {
        String rid = (String) cpitr.next ();
        ArrayList rlist = (ArrayList) this.getProductByReaction ().get (rid);

        if (rlist != null) {
          Iterator cpdtargets = rlist.iterator ();

          while (cpdtargets.hasNext ()) {
            String cpdid2 = (String) cpdtargets.next ();
            int weight2 = (int) ((2000 / this.getCompoundConnectivity (cpdid2))) + 1;

            edges.append (cpdid1 + " -- " + cpdid2 + " [ weight=" + (weight1 + weight2) + " ];");
          }
        }
      }
    }

     return graph + edges.toString () + "}";
  }

  /**
   * loadModifiedWeights permit to modify the weight of a graph nodes with
   * weight loaded from a file the file must be a tab delimited file. First
   * column is the node ID and the second the weight float value
   *
   * @param weightfilename
   * @param graph
   * @return
   * @throws IOException
   */
  public static Graph loadModifiedWeights (String weightfilename, Graph graph)
    throws IOException {

    File weightfile = new File (weightfilename);
    BufferedReader br = null;

    // if(weightfile.exists()){
    br = new BufferedReader (new FileReader (weightfile));

    String sLine = "";
    Hashtable loadedweight = new Hashtable ();

    while ((sLine = br.readLine ()) != null) {
      int pos = sLine.indexOf ('\t');

      if (pos == -1) {
        throw new IOException ("Wrong weight file !!!!");
      }

      loadedweight.put (sLine.substring (0, pos), new Float (sLine.substring (pos + 1)));
    }

    // Graph graph = graphdata.getGraph ();
    ListIterator elementsitr = graph.getElements ().listIterator ();

    while (elementsitr.hasNext ()) {
      Node n = (Node) elementsitr.next ();
      String id = n.getId ();
      Float weight = (Float) loadedweight.get (id);

      if (weight != null) {
        n.getInfo ().setWeight (weight.floatValue ());

        // System.out.println(n.getId() + " " + n.getWeight());
      }
    }

    return graph;
  }

/**
 * @return Returns the modifiedWeights.
 */
public HashMap getModifiedWeights() {
	return modifiedWeights;
}
/**
 * @param modifiedWeights The modifiedWeights to set.
 */
public void setModifiedWeights(HashMap modifiedWeights) {
	this.modifiedWeights = modifiedWeights;
}

public Graph getModifiedGraph() throws Exception{
	if (this.getModifiedWeights() !=null){
		return this.modifyWeigth(this.getGraph(),this.getModifiedWeights());
	} else throw new Exception("No list of modified weights....!");
}
	/**
	 * @param g
	 * @param weights
	 * @return @throws
	 *         RemoteException
	 */
	 public Graph modifyWeigth(Graph g, HashMap weights) {
		ListIterator elementsitr = g.getElements().listIterator();
		HashMap noc = weights;
		while (elementsitr.hasNext()) {
			Node n = (Node) elementsitr.next();
			String id = n.getId();
			if (noc.containsKey(id)) {
				n.getInfo().setWeight(((Float) noc.get(id)).floatValue());
			}
		}
		return g;
	}
}


/*--- Formatted in North Style ---*/
