/*
 * Created on Nov 15, 2006
 *
 * To change this generated comment go to 
 * Window>Preferences>Java>Code Generation>Code Template
 */
package graph.weighted;

import graph.weighted.weightedNodes.Pathfinding;
import graph.weighted.weightedNodes.Result;
import graph.weighted.weightedNodes.Solutions;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Vector;

/**
 * This class launches the pathfinding tool of Fabian Couche and Didier Croes.
 * 
 * @author Karoline Faust
 */
public class PathfindingDidierLauncher {

	// identifier of start node
	private String _startNodeId;
	
	// identifier of end node
	private String _endNodeId;
	
	// maximal weight
	private Float _maxWeight;
	
	// maximal level
	private Integer _maxLevel;
	
	// the number of paths to collect
	private Integer _pathNumber;
	
	// ascii file storing input graph 
	private String _inputGraphFile;
	
	// collect ranks
	private boolean _collectRanks;
		
	// vector of path strings (identifiers in strings separated by pathDelimiter)
	private Vector _pathStringVector;
	
	// timeout in milliseconds
	private long _timeout;
	
	// extended verbosity
	public boolean verbose = false;
	
	// this delimiter is used to append node identifiers to a path string
	public String pathDelimiter = "\t";
	
	/**
	 * Constructor. The constructor receives in this order:
	 * identifier of start node
	 * identifier of end node
	 * maximal weight
	 * maximal level (expressed as number of metabolic steps!)
	 * path number (the number of returned paths)
	 * the name of the ascii file storing the graph
	 * a boolean to flag that ranks should be collected (if true)
	 * @param String startNode
	 * @param String endNode
	 * @param Float maxWeight
	 * @param Integer maxLevel
	 * @param Integer pathNumber
	 * @param String graphFileName
	 * @param boolean collectRanks
	 */
	public PathfindingDidierLauncher(String startNode, String endNode, Float maxWeight, Integer maxLevel, Integer pathNumber, String graphFileName, boolean collectRanks, long timeout){
		this._startNodeId = startNode;
		this._endNodeId = endNode;
		this._maxWeight = maxWeight;
		this._maxLevel = maxLevel;
		this._pathNumber = pathNumber;
		this._collectRanks = collectRanks;
		this._inputGraphFile = graphFileName;
		this._pathStringVector = new Vector();
		this._timeout = timeout;
	}
	
	/**
	 * Collect the GraphData from the given file.
	 * @return GraphData
	 */
	private GraphData collectGraphData(){
		if(verbose)System.out.println(PathfindingDidierLauncher.class.getName()+" Collecting graphData from ascii file.");
		GraphData graphData = new GraphData(this._inputGraphFile,true);
		return graphData;
	}
	
	/**
	 * Print parameters with which pathfinding was launched.
	 */
	private void printParameters(){
		String toPrint = PathfindingDidierLauncher.class.getName()+" Parameters:\n";
		toPrint += "start node identifier: "+this._startNodeId+"\n";
		toPrint += "end node identifier: "+this._endNodeId+"\n";
		toPrint += "maximal weight: "+this._maxWeight+"\n";
		toPrint += "maximal path length (metabolic steps): "+this._maxLevel+"\n";
		toPrint += "requested path number: "+this._pathNumber+"\n";
		toPrint += "collect ranks: "+this._collectRanks+"\n";
		toPrint += "input file: "+this._inputGraphFile+"\n";
		System.out.println(toPrint);
	}
	
	/**
	 * Start pathfinding.
	 */
	public Object launchPathfinding(){
		if(verbose) printParameters();
		// remark: graphData.getGraph returns the weighted graph with compound weights = connectivity and reaction weights = 1
		GraphData graphData = this.collectGraphData();
		Pathfinding pathfinding = new Pathfinding();
		pathfinding.setBacktrackTimeout(this._timeout);
		LinkedList path = null;
		Result res = null;
		String pathString = "";
		try {
			if(verbose)System.out.println(PathfindingDidierLauncher.class.getName()+" Start pathfinding.");
			// List start GraphData gd,int mode (0 = weighted graph), String startID, String endID,
			// int nbSolutions,boolean byrank, float weightM, int levelM, int pathfindingmethod (0 = backtrack)
			List solutions = pathfinding.start(graphData, 0,this._startNodeId,this._endNodeId,this._pathNumber.intValue(),this._collectRanks,this._maxWeight.floatValue(),this._maxLevel.intValue(), 0);
			// get results
			Solutions solutionDisplay = new Solutions();
			solutionDisplay.setPathways(solutions);
			if(verbose)solutionDisplay.textDisplay();
			if(!solutions.isEmpty()){
				for(int i = 0; i < solutions.size(); i++){
					pathString = "";
					res = (Result) solutions.get(i);
					path = res.getPath();
					ListIterator iterator = path.listIterator();
					while(iterator.hasNext()) pathString += iterator.next() + this.pathDelimiter;
					this._pathStringVector.add(pathString);
				}
			}else{
				System.err.println(PathfindingDidierLauncher.class.getName()+" launchPathfinding: No results obtained!");
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return (this._pathStringVector);
	}
		
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// parameters
		String fileName = "ecoliGraph.txt";
		String startID = "R01214";
		String endID = "R00401";
		Float weightM = new Float(500);
		Integer levelM = new Integer(5);
		Integer pathNumber = new Integer(5);
		long timeout = 1800000;
		boolean collectRanks = true;
		// pathfinding 
		PathfindingDidierLauncher launcher = new PathfindingDidierLauncher(startID,endID, weightM, levelM, pathNumber,fileName, collectRanks, timeout);
		launcher.verbose = false;
		Vector paths = (Vector) launcher.launchPathfinding();
		if(!paths.isEmpty()){
			for(int index = 0; index < paths.size(); index++){
				System.out.println(paths.get(index));
			}
		}else{
			System.out.println("Path vector empty!");
		}
	}

}
