package graph.weighted.weightedNodes;
import graph.weighted.GraphData;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;


import EDU.oswego.cs.dl.util.concurrent.TimedCallable;
import EDU.oswego.cs.dl.util.concurrent.TimeoutException;
/**
 * Pathfinding on metabolic networks.
 * 
 * First the backtracking is executed to see if exist the number of solution
 * requested during pathfindingTimeout if exist it stops and the rea is executed
 * if NO solution it stops and return no solutions if it doesn't reach the
 * number of solution in the pathfindingTimeout(default is 10 minute). it stops
 * and the rea tries to do better.
 * 
 * @author Fabian Couche, Didier Croes
 * @version $Id: Pathfinding.java,v 1.25 2005/03/25 16:29:01 didier Exp $
 */
public class Pathfinding {
	
	/** the backtraking timeout in backtrak only mode*/
	private long backtrackTimeout = 3600000;
	
	/** the backtraking timeout in backtrak REA mode*/
	private long backtrackWithREATimeout = 10000;
	/** the REA timeout*/
	private long reaTimeout = 3600000;
	/** whole pathfinding timeout */
	private long pathfindingTimeout = 3600000;
	private List solutionList;
	private Map solutionMap;
	Graph graph = null;
	// //////////////////////////////////////////////////////////
	// 
	// Basic operations
	// 
	// //////////////////////////////////////////////////////////
	/**
	 * Initialises the needed structures with the parameters and launch the
	 * research.
	 * 
	 * @param graphData
	 *            the graph.
	 * @param startID
	 *            The source ID
	 * @param endID
	 *            The destination ID
	 * @param nbSolutions
	 *            the number of solutions.
	 * @param weightM
	 *            The maximum weight
	 * @param levelM
	 *            The maximum level
	 * @return List of Result
	 */
	public synchronized List start(GraphData gd,int mode, String startID, String endID,
			int nbSolutions,boolean byrank, float weightM, int levelM, int pathfindingmethod)
			throws Exception {
		// 
		assert startID != null;
		assert endID != null;
		
		TimedCallable timedCallable;
		List solutions = null;
		int levelMax = 1 + (levelM * 2);
		LOGGER.info("Source: " + startID);
		LOGGER.info("Target: " + endID);
		LOGGER.info("Maximum weight: " + weightM);
		LOGGER.info("Maximum lenght: " + levelMax);
		LOGGER.info("By rank :" + byrank);
		Thread backtrackThread = null;
		if (pathfindingmethod == 2) this.setBacktrackTimeout(this.getBacktrackWithREATimeout());
		else if (pathfindingmethod == 0)this.setBacktrackTimeout(this.getPathfindingTimeout());
		LOGGER.info("TIMEOUT :" + this.getPathfindingTimeout());
		if (pathfindingmethod == 0 || pathfindingmethod == 2) {
			// Launching the backtracking thread
			Backtracking backtrack = new Backtracking();
			backtrack.init(gd, startID, endID,mode, nbSolutions,byrank, weightM,
					levelMax, this.getBacktrackTimeout(), this,
					pathfindingmethod == 0);
			graph = backtrack.getGraph();
			LOGGER.info("BEGIN BACKTRACKING");
			// timedCallable = new TimedCallable (backtrack, 100000);
			backtrackThread = new Thread(backtrack, "Backtracking");
			backtrackThread.start();
			try {
				LOGGER.info("Waiting");
				if (pathfindingTimeout == -1) wait(); // -1 means no timeout
				else wait(pathfindingTimeout);
			
				//throw new Exception("PathFinding timeout");// wait until the
				// backtrack decide to stop or pathfinding Timeout
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			// solutions = (ArrayList) timedCallable.call ();
			LOGGER.info("END BACKTRACKING");
			solutions = backtrack.getSolutions().getPathways();
			
			
		}
		//graph.cleanMarks();
		
		if (solutions == null) LOGGER.warning("backtracking return TIMEOUT!!!!");
		if (solutions == null && pathfindingmethod > 0) { // if the solutions
			// is null means that
			// the backtrakking
			// stops before
			// finishing his job or no backtrakking
			if (backtrackThread != null) {
				backtrackThread.stop(); // Stopping the backtracking thread see
										// if
				// interrup is also OK
				backtrackThread = null;
                LOGGER.warning("Timeout for backtracking... launching REA");
			}else LOGGER.warning("launching REA");
			
			System.gc();
			graph = this.getNewGraph(gd,mode);
			try {
				REA2 rea2 = new REA2();
				//REA rea2 = new REA();
				rea2.setTimeout(this.getReaTimeout());
				timedCallable = new TimedCallable(rea2, this.getReaTimeout());
				rea2
						.init(graph, startID, endID, nbSolutions, byrank,weightM,
								levelMax);
				LOGGER.info("BEGIN REA");
				solutions = (List) timedCallable.call();
				LOGGER.severe("END REA");

			} catch (TimeoutException te) {
				LOGGER.severe("REA has timeout!!!!!");
			} catch (Exception f) {
				LOGGER.severe("It doesn't exist " + nbSolutions
						+ " pathways between these points.");
				f.printStackTrace();
			}
		}
		
		if (solutions != null) {
			
			float sourceHalfWeight = -1;
			float targetHalfWeight = -1;
			int cpdPos = -1;
			LinkedList reacPos = null;
			if ((reacPos = (LinkedList) graph.getReactionPositionsById(startID)) == null) {
				cpdPos = graph.getCompoundPositionById(startID);
				sourceHalfWeight = graph.getNodeByPos(cpdPos).getWeight() / 2;
			} else
				sourceHalfWeight = graph.getNodeByPos(
						((Integer) reacPos.getFirst()).intValue()).getWeight() / 2;
			if ((reacPos = (LinkedList) graph.getReactionPositionsById(endID)) == null) {
				cpdPos = graph.getCompoundPositionById(endID);
				targetHalfWeight = graph.getNodeByPos(cpdPos).getWeight() / 2;
			} else
				targetHalfWeight = graph.getNodeByPos(
						((Integer) reacPos.getFirst()).intValue()).getWeight() / 2;
			Iterator itr = solutions.iterator();
			while (itr.hasNext()) {
				Result res = (Result) itr.next();
				res.setDistance(res.getScore() - sourceHalfWeight
						- targetHalfWeight);
			}
		}
		
		//System.out.println("Rank NB = " + this.getSolutionsMap().size());
		return solutions;
	}
	/**
	 *  
	 */
	public synchronized void notifyBacktrackEnd() {
		this.notify();
	}
	/**
	 * @param level
	 */
	public static void setMsgLevel(Level level) {
		// LOGGER.setLevel (level);
	}
	// //////////////////////////////////////////////////////////
	// 
	// Implementation
	// 
	// //////////////////////////////////////////////////////////
	/** Field LOGGER */
	private static final Logger LOGGER = Logger.getLogger(Pathfinding.class
			.getName());
	/**
	 * @return
	 */
	public long getBacktrackWithREATimeout() {
		return backtrackWithREATimeout;
	}
	/**
	 * @return
	 */
	public long getReaTimeout() {
		return reaTimeout;
	}
	/**
	 * @param l
	 */
	public void setBacktrackWithREATimeout(long l) {
		backtrackWithREATimeout = l;
	}
	/**
	 * @param l
	 */
	public void setReaTimeout(long l) {
		reaTimeout = l;
	}
	/**
	 * @return Returns the pathfindingTimeout.
	 */
	public long getPathfindingTimeout() {
		return pathfindingTimeout;
	}
	/**
	 * @param pathfindingTimeout The pathfindingTimeout to set.
	 */
	public void setPathfindingTimeout(long pathfindingTimeout) {
		if (pathfindingTimeout != -2) // -2 means keep the default value
		this.pathfindingTimeout = pathfindingTimeout;
	}
	/**
	 * @return Returns the backtrackTimeout.
	 */
	public long getBacktrackTimeout() {
		return backtrackTimeout;
	}
	/**
	 * @param backtrackTimeout The backtrackTimeout to set.
	 */
	public void setBacktrackTimeout(long backtrackTimeout) {
		this.backtrackTimeout = backtrackTimeout;
	}
	public Graph getNewGraph(GraphData graphdata, int PFMethod ) throws Exception{
		switch (PFMethod) {// get the graph requested
		case 0 :
			return graphdata.getGraph();
			
		case 1 :
			return graphdata.getGraph(1);
		
		case 2 :
			return graphdata.getGraph(1);	
		default :
			throw new Exception("Not a valid Mode!");
		
	// break;
	}
		
	}
	/**
	 * @return Returns the graph.
	 */
	public Graph getGraph() {
		return graph;
	}
}
/*--- Formatted in North Style ---*/
