package graph.weighted.weightedNodes;



import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.TreeMap;

import EDU.oswego.cs.dl.util.concurrent.Callable;


/**
 * Recursive Enumeration Algorithm.
 *
 * @author Fabian Couche
 * @version $Id: REA2.java,v 1.8 2005/03/25 16:29:01 didier Exp $
 */
public class REA2
  implements Callable, Runnable {

  // //////////////////////////////////////////////////////////
  // 
  // Construction
  // 
  // //////////////////////////////////////////////////////////

  /**
   * REA Initialization
   * @param graph the graph to use.
   * @param nb the number of solutions.
   * @param startID the id of the start node.
   * @param endID the id of the end node.
   * @param weightM the max weight.
   * @param levelM the max level.
   */
  public void init (Graph graph, String startID, String endID, int nb, boolean byrank,float weightM, int levelM) {

    this.setGraph (graph);
    
    _start = startID;
    _end = endID;
    _solutions = new ArrayList ();
    _tempsolutionsmap = new TreeMap();
    _byRank= byrank;
    _currentWay = new LinkedList ();
    _limit = weightM;
    _levelMax = levelM;
    _nbSolutions = nb;
    _timeOver = false;
    this._isInitialized = true;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Sets the graph to use.
   * @param graph the current graph.
   */
  protected void setGraph (Graph graph) {
    _graph = graph;
  }

  /**
   * Returns the used graph.
   * @return the current graph.
   */
  protected Graph getGraph () {
    return _graph;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Basic operations
  // 
  // //////////////////////////////////////////////////////////


  /**
   * Is Called when the REA is launched as a thread.
   * @return an Object.
   * @throws Exception a exception if it was not initialized.
   */
  public Object call ()
    throws Exception {

    if (!this._isInitialized) {
      throw new Exception ("REA was not initialized!");
    } else {

      if (!this.getGraph ().containsCompoundId (_start)) {
        
        LinkedList couple = this.getGraph ().getReactionPositionsById (_start);
        
        Integer startPos1 = (Integer) couple.getFirst ();
        Integer startPos2 = (Integer) couple.getLast ();
        
        Node startNode1, startNode2;
        
        startNode1 = (Node) this.getGraph ().getNodeByPos (startPos1.intValue ());
        startNode2 = (Node) this.getGraph ().getNodeByPos (startPos2.intValue ());

        _currentWay.add (0, startNode1.getId ());

        _currentDistance = startNode1.getWeight ();

        rea2 (startNode1, startNode2);
      } else {
        int startPos = this.getGraph ().getCompoundPositionById (_start);
        Node startNode = (Node) this.getGraph ().getNodeByPos (startPos);

        _currentWay.add (0, startNode.getId ());

        _currentDistance = startNode.getWeight ();

        this.rea2 (startNode);
      }
    }
    if (_byRank){
		
		this.setSolutionsMap(_tempsolutionsmap);
		_solutions = new ArrayList();
		Iterator itr = this.getSolutionsMap().values().iterator();
		while(itr.hasNext()){
			_solutions.addAll((ArrayList)itr.next());
		}
	
}else {
		this.setSolutionsMap(null);
}
    return _solutions;
   }

  /**
   * version of REA that uses Dijkstra with constraints
   * @param startNode A Node
   */
  public void rea2 (Node startNode) {

    // 
    assert startNode != null;

    // 
    boolean accept = true;
    int t = this.getGraph ().getCompoundPositionById (_end);
    int n = _nbSolutions;
    long inTime;

    inTime = System.currentTimeMillis ();

    Path last;

    this.dijkstraWithConstraints (startNode);
    
    accept = this.buildPath (t, 0);
    System.out.println("dijkstra = "+_currentDistance);

    if (!accept) {
      n++;
    }

    _c = new LinkedList [this.getGraph ().getNumberElements ()];
    this._c [_source] = new LinkedList ();

    for (int k = 1; k < n; ++k) {
    	
      this.nextPath (t, k);
    
      accept = buildPath (t, k);
      System.out.println(_currentDistance);
      if (!accept) {
        n++;
      }

      last = (Path) _pi [t].get (k);

      if (last.getBack () == null) {
        k += n;
      }

      if (System.currentTimeMillis () - inTime >= _timeout) {
        _timeOver = true;
        k += n;
      }
    }
  }

  /**
   * An adapted version of the famous Dijkstra algorithm with constraints
   * @param startNode A Node
   */
  public void dijkstraWithConstraints (Node startNode) {

    // 
    assert startNode != null;

    // 
    Integer o;  // variable of work
    Node n;     // variable of work
    Path k, s;
    int taille = this.getGraph ().getNumberElements ();
    int cpt = taille - 1;
    int startPos = this.getGraph ().getNodePosition (startNode);

    _source = startPos;

    Integer startPosition = new Integer (startPos);
    Set m = new HashSet ();
    ArrayList fringe = new ArrayList ();
    float dist;
    boolean insertion;

    // INITILISATION OF THE STRUCTURES
    _pi = new LinkedList [taille];
    _pi [startPos] = new LinkedList ();
    s = new Path (startPos, 0, startPos);

    _pi [startPos].add (s);
    m.add (startPosition);
    this.insert (s, fringe);

    for (int i = 0; i < taille; ++i) {
      if (i != startPos) {
        o = new Integer (i);
        insertion = false;
        n = (Node) this.getGraph ().getNodeByPos (i);

        if ((this.existLink (startNode, i))) {
          dist = n.getWeight ();
          insertion = true;
        } else {
          dist = Integer.MAX_VALUE;
        }

        m.add (o);

        _pi [i] = new LinkedList ();
        k = new Path (s, dist, i);

        _pi [i].add (k);

        if ((insertion) && this.verifyConstraints (s, n)) {
          insert (k, fringe);
        }
      }
    }

    m.remove (startPosition);
    fringe.remove (startPosition);

    // BEGINNING OF THE SEARCH
    boolean finish = false;
    Integer minId;
    Path minPath = null;
    float minVal;

    while (!finish) {

      // CHOICE OF THE NEXT NODE TO INSERT IN SOLUTION
      minId = null;
      minVal = Integer.MAX_VALUE;
      minPath = (Path) fringe.get (0);
      minVal = minPath.getLenght ();
      minId = new Integer (minPath.getNode ());

      if (minVal == Integer.MAX_VALUE) {
        return;
      }

      m.remove (minId);
      fringe.remove (minPath);

      cpt--;

      // ANALYSE FOLLOWERS OF THE SELECTED NODE
      n = (Node) this.getGraph ().getNodeByPos (minId.intValue ());

      for (ListIterator i = n.getNxt ().listIterator (); i.hasNext (); ) {
        Integer f = (Integer) i.next ();
        int futur = f.intValue ();

        if (m.contains (f)) {
          Node futurNode = (Node) this.getGraph ().getNodeByPos (futur);
          float v = minVal + futurNode.getWeight ();

          k = (Path) _pi [futur].getFirst ();

          if ((v < k.getLenght ()) && this.verifyConstraints (minPath, futurNode)) {
            k.setLenght (v);
            k.setBack (minPath);

            if (fringe.contains (k)) {
              fringe.remove (k);
            }

            insert (k, fringe);
          }
        }
      }

      if (fringe.isEmpty ()) {
        finish = true;
      }
    }
  }

  /**
   * Inserts in right position a Path in a list.
   * @param k a path.
   * @param l an arrayList.
   */
  public void insert (Path k, ArrayList l) {

    // 
    assert k != null;

    assert l != null;

    // 
    Path last;
    int n = l.size ();
    int pos = 0;
    int inf = 0;
    int sup = n;

    while (sup - inf >= 1) {
      last = (Path) l.get (pos);

      if (k.getLenght () < last.getLenght ()) {
        sup = pos;
      } else {
        inf = pos + 1;
      }

      pos = (inf + sup) / 2;
    }

    l.add (pos, k);
  }

  /**
   * Verifies if the node has an outgoing link to the index position in the graph
   * @param n A node
   * @param pos A position in the index of the graph
   * @return a boolean
   */
  public boolean existLink (Node n, int pos) {

    // 
    assert n != null;

    // 
    int intValue;
    Integer integerValue;
    boolean found = false;

    for (ListIterator x = n.getNxt ().listIterator (); x.hasNext (); ) {
      integerValue = (Integer) x.next ();
      intValue = integerValue.intValue ();

      if (intValue == pos) {
        found = true;

        break;
      }
    }

    return found;
  }

  /**
   * Verifies the constraints
   * @param p A Path
   * @param n A Node
   * @return a boolean
   */
  public boolean verifyConstraints (Path p, Node n) {

    // 
    assert p != null;

    assert n != null;

    // 
    int cpt = 1;
    boolean test;
    boolean result = false;
    Node elem;
    String id = n.getId ();
    float val = p.getLenght ();

    val += n.getWeight ();

    if (val < _limit) {
      do {
        test = true;
        elem = (Node) this.getGraph ().getNodeByPos (p.getNodeId ());

        if (elem.getId () == id) {
          test = false;
        }

        p = p.getBack ();

        cpt++;
      } while ((p.getBack () != p) && (cpt < _levelMax) && (test));

      if (p.getBack () == p) {
        result = true;
      }
    }

    return result;
  }

  /**
   * nextPath Method
   * @param v An int
   * @param k An int
   */
  public void nextPath (int v, int k) {

    Integer integerVal;
    Node n;
    int intVal, kp;
    Path p, temp;

    if (v != this._source) {
      if (k == 1) {
        n = (Node) this.getGraph ().getNodeByPos (v);
        this._c [v] = new LinkedList ();

        for (ListIterator i = n.getPrev ().listIterator (); i.hasNext (); ) {
          integerVal = (Integer) i.next ();
          intVal = integerVal.intValue ();
          p = (Path) _pi [intVal].getFirst ();
          temp = (Path) _pi [v].getFirst ();

          if (p != temp.getBack ()) {
            _c [v].add (p);
          }
        }
      }

      temp = (Path) _pi [v].get (k - 1);
      p = temp.getBack ();

      if (p != null) {
        intVal = p.getNode ();
        kp = _pi [intVal].indexOf (p);

        if ((kp + 1) == _pi [intVal].size ()) {
          nextPath (intVal, kp + 1);
        }

        temp = (Path) _pi [intVal].get (kp + 1);

        if (temp.getBack () != null) {
          _c [v].add (temp);
        }
      }
      
    }

    if (_c [v].size () != 0) {
      Path min = (Path) _c [v].getFirst ();
      float l;

      for (ListIterator i = _c [v].listIterator (); i.hasNext (); ) {
        temp = (Path) i.next ();

        if (temp.getLenght () < min.getLenght ()) {
          min = temp;
        }
      }

      l = min.getLenght ();
      n = (Node) this.getGraph ().getNodeByPos (v);
      l += n.getWeight ();
      temp = new Path (min, l, v);
      
      _pi [v].add (temp);
      _c [v].remove (min);
    } else {
      temp = new Path (v);

      _pi [v].add (temp);
    }
    
  }

  /**
   * buildPath Method
   * @param target the destination position.
   * @param k the researched path.
   * @return a boolean
   */
  public boolean buildPath (int target, int k) {

    Node elem;
    LinkedList marked = new LinkedList ();
    int posit;
    int depht = 0;

    _currentWay = new LinkedList ();

    Path p;

    p = (Path) _pi [target].get (k);
    _currentDistance = p.getLenght ();
    
    if (p.getLenght () < _limit) {
      do {
        posit = p.getNode ();
        elem = (Node) this.getGraph ().getNodeByPos (posit);

        if (elem.marked ()) {
          this.removeMark (marked);

          return false;
        } else {
          elem.mark ();
          marked.add (elem);
          _currentWay.addFirst (elem.getId ());

          p = p.getBack ();

          depht++;
        }

        if (depht > _levelMax) {
          this.removeMark (marked);

          return false;
        }
      } while (p.getBack () != p);

      posit = p.getNode ();
      elem = (Node) this.getGraph ().getNodeByPos (posit);

      if (!elem.marked ()) {
        _currentWay.addFirst (elem.getId ());

        _currentDistance += elem.getWeight ();
      } else {
        this.removeMark (marked);

        return false;
      }

      this.updateSolution (_byRank);
      this.removeMark (marked);

      return true;
    }

    return false;
  }

  /**
   * removeMark Method
   * @param marked the list to unmark.
   */
  public void removeMark (LinkedList marked) {

    Node n;

    for (ListIterator i = marked.listIterator (); i.hasNext (); ) {
      n = (Node) i.next ();

      n.unmark ();
    }
  }
  
  public void printListSize(){
  	try{
  	  	System.out.print("c=" + _c.length +" pi=" + _pi.length);
  	  	}catch(NullPointerException npe ){
  	  		System.out.print("c=null pi=null");
  	  	}
  	  	try {
  	  	System.out.print(" c2=" + _c2.length +" pi2=" + _pi2.length);
  	  	}catch(NullPointerException npe ){
  	  		System.out.print("c2=null pi2=null");
  	  	}
  	  	try {
  	  	System.out.print(" c3=" + _c3.length +" pi3=" + _pi3.length);
  	  	}catch(NullPointerException npe ){
  	  		System.out.print("c3=null pi3=null");
  	  	}
  	  	try {
  	  	System.out.println(" c4=" + _c4.length +" pi4=" + _pi4.length);
  	  	}catch(NullPointerException npe ){
  	  		System.out.print("c4=null pi4=null");
  	  	}
  }
  private void updateSolution(boolean byRank){
  	if (byRank) this.updateSolutionKRank();
  	else this.updateSolutionKShortest() ;
  	//if (_nbSolutions == _solutions.size()) finnished = true;
  }
  /**
   *  Used by the algorithm of pathfinding to update
   *  the list of solutions. It works with the global structures and does a
   *  dichotomic insertion. The limit of the number of solutions is considered
   *  in the end of this procedure (the worstsolution is removed).
   */
  public void updateSolutionKShortest () {
    
    Result res = new Result (_currentWay, _currentDistance);
    Result last;
    int n = _solutions.size ();
    int pos = 0;
    int inf = 0;
    int sup = n;
    while (sup - inf >= 1) {
      last = (Result) _solutions.get (pos);
    
      if (_currentDistance < last.getScore ()) {
        sup = pos;
      } else {
        inf = pos + 1;
      }

      pos = (inf + sup) / 2;
    }

    _solutions.add (pos, res);

    n++;

    if (n > _nbSolutions) {
      _solutions.remove (n - 1);

      last = (Result) _solutions.get (n - 2);
      _limit = last.getScore ();
    }
  }
  private void updateSolutionKRank () {

    Result res = new Result (this._currentWay, this._currentDistance);
    Result last;
    
    Float n = null;
    Float currentdistance = new Float(this._currentDistance);
    ArrayList pathwaylist = null;
    if (!this._tempsolutionsmap.isEmpty()){
    		n = (Float) this._tempsolutionsmap.lastKey();
   		pathwaylist = (ArrayList) this._tempsolutionsmap.get(currentdistance);
    }

    if(pathwaylist == null){
    	if (n == null || n.compareTo(currentdistance) > 0) {
    		pathwaylist = new ArrayList();
    		pathwaylist.add(res);
    		this._tempsolutionsmap.put(currentdistance,pathwaylist);
    		if(this._tempsolutionsmap.size() > this._nbSolutions){
    			this._tempsolutionsmap.remove(this._tempsolutionsmap.lastKey());
    		
    		}
    		_limit =  (((Float)this._tempsolutionsmap.lastKey()).floatValue());

    	    }
    }else{
    		pathwaylist.add(res);
    }
    
    

    
  }
////////////////////////////////////////////////////////////
//
// Research between reactions
//
////////////////////////////////////////////////////////////

  /**
   * REA2 Method
   * @param startNode1 the first node.
   * @param startNode2 the second node.
   */
  public void rea2 (Node startNode1, Node startNode2) {

    // 
    assert startNode1 != null;

    assert startNode2 != null;

    // 
    boolean accept = true;
    LinkedList couple = (LinkedList) this.getGraph().getReactionPositionsById (_end);
    Integer endPosition1 = (Integer) couple.getFirst ();
    Integer endPosition2 = (Integer) couple.getLast ();
    int endPos1 = endPosition1.intValue ();
    int endPos2 = endPosition2.intValue ();
    int n = _nbSolutions;
    int k1 = 1;
    int k2 = 1;
    int k3 = 1;
    int k4 = 1;
    int min;
    float minVal;
    float niveau [] = new float [4];
    long inTime;

    inTime = System.currentTimeMillis ();

    Path last, p;

    last = null;
    n += 3;

    this.dijkstraWithConstraints (startNode1);

    accept = this.buildPath (endPos1, 0);

    if (!accept) {
      n++;
    }

    this.dijkstraWithConstraints2 (startNode2);

    accept = this.buildPath2 (endPos1, 0);

    if (!accept) {
      n++;
    }

    this.dijkstraWithConstraints3 (startNode1);

    accept = this.buildPath3 (endPos2, 0);

    if (!accept) {
      n++;
    }

    this.dijkstraWithConstraints4 (startNode2);

    accept = this.buildPath4 (endPos2, 0);

    if (!accept) {
      n++;
    }

    _c = new LinkedList [this.getGraph ().getNumberElements ()];
    this._c [_source] = new LinkedList ();
    _c2 = new LinkedList [this.getGraph ().getNumberElements ()];
    this._c2 [_source2] = new LinkedList ();
    _c3 = new LinkedList [this.getGraph ().getNumberElements ()];
    this._c3 [_source3] = new LinkedList ();
    _c4 = new LinkedList [this.getGraph ().getNumberElements ()];
    this._c4 [_source4] = new LinkedList ();

    while ((k1 + k2 + k3 + k4) < n) {
    	 
      p = (Path) _pi [endPos1].get (k1 - 1);
      niveau [0] = p.getLenght ();
      p = (Path) _pi2 [endPos1].get (k2 - 1);
      niveau [1] = p.getLenght ();
      p = (Path) _pi3 [endPos2].get (k3 - 1);
      niveau [2] = p.getLenght ();
      p = (Path) _pi4 [endPos2].get (k4 - 1);
      niveau [3] = p.getLenght ();
      minVal = niveau [0];
      min = 0;
    
      for (int i = 1; i < 4; i++) {
        if (niveau [i] < minVal) {
          minVal = niveau [i];
          min = i;
          
        }
      }
      
      //if (_solutions.size () == _nbSolutions && _limit <= minVal ) break; // added by didier stop the rea 
      switch (min) {

        case 0 :
          nextPath (endPos1, k1);

          accept = buildPath (endPos1, k1);
          last = (Path) _pi [endPos1].get (k1);

          k1++;
          break;

        case 1 :
          this.nextPath2 (endPos1, k2);

          accept = this.buildPath2 (endPos1, k2);
          last = (Path) _pi2 [endPos1].get (k2);

          k2++;
          break;

        case 2 :
          this.nextPath3 (endPos2, k3);

          accept = this.buildPath3 (endPos2, k3);
          last = (Path) _pi3 [endPos2].get (k3);

          k3++;
          break;

        case 3 :
          this.nextPath4 (endPos2, k4);

          accept = this.buildPath4 (endPos2, k4);
          last = (Path) _pi4 [endPos2].get (k4);

          k4++;
          break;
      }

      if (!accept) {
        n++;
      }else
      System.out.println("\ta = " +accept);
      if (last.getBack () == null) {
        k1 += n;
      }

      if (System.currentTimeMillis () - inTime >= _timeout) {
        _timeOver = true;
        k1 += n;
      }
     // if (finnished) break;
    }

    // System.out.println("Total number: "+n);
  }

  /**
   * dijkstraWithConstraints4 Method
   * @param startNode the start point.
   */
  public void dijkstraWithConstraints4 (Node startNode) {

    // 
    assert startNode != null;

    // 
    Integer o;  // variable of work
    Node n;     // variable of work
    Path k, s;
    int taille = this.getGraph ().getNumberElements ();
    int cpt = taille - 1;
    int startPos = this.getGraph ().getNodePosition (startNode);

    _source4 = startPos;

    Integer startPosition = new Integer (startPos);
    Set m = new HashSet ();
    ArrayList fringe = new ArrayList ();
    long inTime, outTime;
    long initTime = 0;
    long minTime = 0;
    long succTime = 0;
    float dist;
    boolean insertion;

    // INITILISATION OF THE STRUCTURES
    inTime = System.currentTimeMillis ();
    _pi4 = new LinkedList [taille];
    _pi4 [startPos] = new LinkedList ();
    s = new Path (startPos, 0, startPos);

    _pi4 [startPos].add (s);
    m.add (startPosition);
    insert (s, fringe);

    for (int i = 0; i < taille; ++i) {
      if (i != startPos) {
        o = new Integer (i);
        insertion = false;
        n = (Node) this.getGraph ().getNodeByPos (i);

        if ((existLink (startNode, i))) {
          dist = n.getWeight ();
          insertion = true;
        } else {
          dist = Integer.MAX_VALUE;
        }

        m.add (o);

        _pi4 [i] = new LinkedList ();
        k = new Path (s, dist, i);

        _pi4 [i].add (k);

        if ((insertion) && verifyConstraints (s, n)) {
          insert (k, fringe);
        }
      }
    }

    m.remove (startPosition);
    fringe.remove (startPosition);

    outTime = System.currentTimeMillis ();
    initTime = outTime - inTime;

    System.out.println ("Temps d'initialisation: " + initTime);

    // BEGINNING OF THE SEARCH
    boolean finish = false;
    Integer minId;
    Path minPath = null;
    float minVal;

    while (!finish) {

      // CHOICE OF THE NEXT NODE TO INSERT IN SOLUTION
      inTime = System.currentTimeMillis ();
      minId = null;
      minVal = Integer.MAX_VALUE;
      minPath = (Path) fringe.get (0);
      minVal = minPath.getLenght ();
      minId = new Integer (minPath.getNode ());

      if (minVal == Integer.MAX_VALUE) {
        return;
      }

      m.remove (minId);
      fringe.remove (minPath);

      outTime = System.currentTimeMillis ();
      minTime += (outTime - inTime);

      cpt--;

      // ANALYSE FOLLOWERS OF THE SELECTED NODE
      inTime = System.currentTimeMillis ();
      n = (Node) this.getGraph ().getNodeByPos (minId.intValue ());

      for (ListIterator i = n.getNxt ().listIterator (); i.hasNext (); ) {
        Integer f = (Integer) i.next ();
        int futur = f.intValue ();

        if (m.contains (f)) {
          Node futurNode = (Node) this.getGraph ().getNodeByPos (futur);
          float v = minVal + futurNode.getWeight ();

          k = (Path) _pi4 [futur].getFirst ();

          if ((v < k.getLenght ()) && verifyConstraints (minPath, futurNode)) {
            k.setLenght (v);
            k.setBack (minPath);

            if (fringe.contains (k)) {
              fringe.remove (k);
            }

            insert (k, fringe);
          }
        }
      }

      outTime = System.currentTimeMillis ();
      succTime += (outTime - inTime);

      if (fringe.isEmpty ()) {
        finish = true;

        System.out.println ("Temps de recherche du min: " + minTime);
        System.out.println ("Temps de traitement des successeurs: " + succTime);
      }
    }
  }


  /**
   * dijkstraWithConstraints3 Method
   * @param startNode the start point.
   */
  public void dijkstraWithConstraints3 (Node startNode) {

    // 
    assert startNode != null;

    // 
    Integer o;  // variable of work
    Node n;     // variable of work
    Path k, s;
    int taille = this.getGraph ().getNumberElements ();
    int cpt = taille - 1;
    int startPos = this.getGraph ().getNodePosition (startNode);

    _source3 = startPos;

    Integer startPosition = new Integer (startPos);
    Set m = new HashSet ();
    ArrayList fringe = new ArrayList ();
    long inTime, outTime;
    long initTime = 0;
    long minTime = 0;
    long succTime = 0;
    float dist;
    boolean insertion;

    // INITILISATION OF THE STRUCTURES
    inTime = System.currentTimeMillis ();
    _pi3 = new LinkedList [taille];
    _pi3 [startPos] = new LinkedList ();
    s = new Path (startPos, 0, startPos);

    _pi3 [startPos].add (s);
    m.add (startPosition);
    insert (s, fringe);

    for (int i = 0; i < taille; ++i) {
      if (i != startPos) {
        o = new Integer (i);
        insertion = false;
        n = (Node) this.getGraph ().getNodeByPos (i);

        if ((existLink (startNode, i))) {
          dist = n.getWeight ();
          insertion = true;
        } else {
          dist = Integer.MAX_VALUE;
        }

        m.add (o);

        _pi3 [i] = new LinkedList ();
        k = new Path (s, dist, i);

        _pi3 [i].add (k);

        if ((insertion) && verifyConstraints (s, n)) {
          insert (k, fringe);
        }
      }
    }

    m.remove (startPosition);
    fringe.remove (startPosition);

    outTime = System.currentTimeMillis ();
    initTime = outTime - inTime;

    System.out.println ("Temps d'initialisation: " + initTime);

    // BEGINNING OF THE SEARCH
    boolean finish = false;
    Integer minId;
    Path minPath = null;
    float minVal;

    while (!finish) {

      // CHOICE OF THE NEXT NODE TO INSERT IN SOLUTION
      inTime = System.currentTimeMillis ();
      minId = null;
      minVal = Integer.MAX_VALUE;
      minPath = (Path) fringe.get (0);
      minVal = minPath.getLenght ();
      minId = new Integer (minPath.getNode ());

      if (minVal == Integer.MAX_VALUE) {
        return;
      }

      m.remove (minId);
      fringe.remove (minPath);

      outTime = System.currentTimeMillis ();
      minTime += (outTime - inTime);

      cpt--;

      // ANALYSE FOLLOWERS OF THE SELECTED NODE
      inTime = System.currentTimeMillis ();
      n = (Node) this.getGraph ().getNodeByPos (minId.intValue ());

      for (ListIterator i = n.getNxt ().listIterator (); i.hasNext (); ) {
        Integer f = (Integer) i.next ();
        int futur = f.intValue ();

        if (m.contains (f)) {
          Node futurNode = (Node) this.getGraph ().getNodeByPos (futur);
          float v = minVal + futurNode.getWeight ();

          k = (Path) _pi3 [futur].getFirst ();

          if ((v < k.getLenght ()) && verifyConstraints (minPath, futurNode)) {
            k.setLenght (v);
            k.setBack (minPath);

            if (fringe.contains (k)) {
              fringe.remove (k);
            }

            insert (k, fringe);
          }
        }
      }

      outTime = System.currentTimeMillis ();
      succTime += (outTime - inTime);

      if (fringe.isEmpty ()) {
        finish = true;

        System.out.println ("Temps de recherche du min: " + minTime);
        System.out.println ("Temps de traitement des successeurs: " + succTime);
      }
    }
  }


  /**
   * dijkstraWithConstraints2 Method
   * @param startNode the start point.
   */
  public void dijkstraWithConstraints2 (Node startNode) {

    // 
    assert startNode != null;

    // 
    Integer o;  // variable of work
    Node n;     // variable of work
    Path k, s;
    int taille = this.getGraph ().getNumberElements ();
    int cpt = taille - 1;
    int startPos = this.getGraph ().getNodePosition (startNode);

    _source2 = startPos;

    Integer startPosition = new Integer (startPos);

    // Integer endPosition= (Integer) compounds.get(end);
    // int endPos= endPosition.intValue();
    Set m = new HashSet ();
    ArrayList fringe = new ArrayList ();
    long inTime, outTime;
    long initTime = 0;
    long minTime = 0;
    long succTime = 0;
    float dist;
    boolean insertion;

    // INITILISATION OF THE STRUCTURES
    inTime = System.currentTimeMillis ();
    _pi2 = new LinkedList [taille];
    _pi2 [startPos] = new LinkedList ();
    s = new Path (startPos, 0, startPos);

    _pi2 [startPos].add (s);
    m.add (startPosition);
    insert (s, fringe);

    for (int i = 0; i < taille; ++i) {
      if (i != startPos) {
        o = new Integer (i);
        insertion = false;
        n = (Node) this.getGraph ().getNodeByPos (i);

        if ((existLink (startNode, i))) {
          dist = n.getWeight ();
          insertion = true;
        } else {
          dist = Integer.MAX_VALUE;
        }

        m.add (o);

        _pi2 [i] = new LinkedList ();
        k = new Path (s, dist, i);

        _pi2 [i].add (k);

        if ((insertion) && verifyConstraints (s, n)) {
          insert (k, fringe);
        }
      }
    }

    m.remove (startPosition);
    fringe.remove (startPosition);

    outTime = System.currentTimeMillis ();
    initTime = outTime - inTime;

    System.out.println ("Temps d'initialisation: " + initTime);

    // BEGINNING OF THE SEARCH
    boolean finish = false;
    Integer minId;
    Path minPath = null;
    float minVal;

    while (!finish) {

      // CHOICE OF THE NEXT NODE TO INSERT IN SOLUTION
      inTime = System.currentTimeMillis ();
      minId = null;
      minVal = Integer.MAX_VALUE;
      minPath = (Path) fringe.get (0);
      minVal = minPath.getLenght ();
      minId = new Integer (minPath.getNode ());

      if (minVal == Integer.MAX_VALUE) {
        return;
      }

      m.remove (minId);
      fringe.remove (minPath);

      outTime = System.currentTimeMillis ();
      minTime += (outTime - inTime);

      cpt--;

      // ANALYSE FOLLOWERS OF THE SELECTED NODE
      inTime = System.currentTimeMillis ();
      n = (Node) this.getGraph ().getNodeByPos (minId.intValue ());

      for (ListIterator i = n.getNxt ().listIterator (); i.hasNext (); ) {
        Integer f = (Integer) i.next ();
        int futur = f.intValue ();

        if (m.contains (f)) {
          Node futurNode = (Node) this.getGraph ().getNodeByPos (futur);
          float v = minVal + futurNode.getWeight ();

          k = (Path) _pi2 [futur].getFirst ();

          if ((v < k.getLenght ()) && verifyConstraints (minPath, futurNode)) {
            k.setLenght (v);
            k.setBack (minPath);

            if (fringe.contains (k)) {
              fringe.remove (k);
            }

            insert (k, fringe);
          }
        }
      }

      outTime = System.currentTimeMillis ();
      succTime += (outTime - inTime);

      if (fringe.isEmpty ()) {
        finish = true;

        System.out.println ("Temps de recherche du min: " + minTime);
        System.out.println ("Temps de traitement des successeurs: " + succTime);
      }
    }
  }

  /**
   * buildPath4 Method
   * @param target the destination node.
   * @param k the researched path.
   * @return the status.
   * @return
   */
  public boolean buildPath4 (int target, int k) {

    Node elem;
    LinkedList marked = new LinkedList ();
    int posit;
    int depht = 0;

    _currentWay = new LinkedList ();

    Path p;

    p = (Path) _pi4 [target].get (k);
    _currentDistance = p.getLenght ();

    if (p.getLenght () < _limit) {
      do {
        posit = p.getNode ();
        elem = (Node) this.getGraph ().getNodeByPos (posit);

        if (elem.marked ()) {
          removeMark (marked);

          return false;
        } else {
          elem.mark ();
          marked.add (elem);
          _currentWay.addFirst (elem.getId ());

          p = p.getBack ();

          depht++;
        }

        if (depht > _levelMax) {
          removeMark (marked);

          return false;
        }
      } while (p.getBack () != p);

      posit = p.getNode ();
      elem = (Node) this.getGraph ().getNodeByPos (posit);

      if (!elem.marked ()) {
        _currentWay.addFirst (elem.getId ());

        _currentDistance += elem.getWeight ();
      } else {
        removeMark (marked);

        return false;
      }

      updateSolution (_byRank);
      removeMark (marked);

      return true;
    }

    return false;
  }


  /**
   * buildPath3 Method
   * @param target the destination node.
   * @param k the researched path.
   * @return the status.
   */
  public boolean buildPath3 (int target, int k) {

    Node elem;
    LinkedList marked = new LinkedList ();
    int posit;
    int depht = 0;

    _currentWay = new LinkedList ();

    Path p;

    p = (Path) _pi3 [target].get (k);
    _currentDistance = p.getLenght ();

    if (p.getLenght () < _limit) {
      do {
        posit = p.getNode ();
        elem = (Node) this.getGraph ().getNodeByPos (posit);

        if (elem.marked ()) {
          removeMark (marked);

          return false;
        } else {
          elem.mark ();
          marked.add (elem);
          _currentWay.addFirst (elem.getId ());

          p = p.getBack ();

          depht++;
        }

        if (depht > _levelMax) {
          removeMark (marked);

          return false;
        }
      } while (p.getBack () != p);

      posit = p.getNode ();
      elem = (Node) this.getGraph ().getNodeByPos (posit);

      if (!elem.marked ()) {
        _currentWay.addFirst (elem.getId ());

        _currentDistance += elem.getWeight ();
      } else {
        removeMark (marked);

        return false;
      }

      updateSolution (_byRank);
      removeMark (marked);

      return true;
    }

    return false;
  }


  /**
   * buildPath2 Method
   * @param target the destination node.
   * @param k the researched path.
   * @return the status.
   */
  public boolean buildPath2 (int target, int k) {

    Node elem;
    LinkedList marked = new LinkedList ();
    int posit;
    int depht = 0;

    _currentWay = new LinkedList ();

    Path p;

    p = (Path) _pi2 [target].get (k);
    _currentDistance = p.getLenght ();

    if (p.getLenght () < _limit) {
      do {
        posit = p.getNode ();
        elem = (Node) this.getGraph ().getNodeByPos (posit);

        if (elem.marked ()) {
          removeMark (marked);

          return false;
        } else {
          elem.mark ();
          marked.add (elem);
          _currentWay.addFirst (elem.getId ());

          p = p.getBack ();

          depht++;
        }

        if (depht > _levelMax) {
          removeMark (marked);

          return false;
        }
      } while (p.getBack () != p);

      posit = p.getNode ();
      elem = (Node) this.getGraph ().getNodeByPos (posit);

      if (!elem.marked ()) {
        _currentWay.addFirst (elem.getId ());

        _currentDistance += elem.getWeight ();
      } else {
        removeMark (marked);

        return false;
      }

      updateSolution (_byRank);

      // System.out.println(k+"th path ACCEPTED");
      removeMark (marked);

      return true;
    }

    return false;
  }


  /**
   * nextPath4 Method
   * @param v the node.
   * @param k the current position.
   */
  public void nextPath4 (int v, int k) {

    Integer integerVal;
    Node n;
    int intVal, kp;
    Path p, temp;

    if (v != this._source4) {
      if (k == 1) {
        n = (Node) this.getGraph ().getNodeByPos (v);
        this._c4 [v] = new LinkedList ();

        for (ListIterator i = n.getPrev ().listIterator (); i.hasNext (); ) {
          integerVal = (Integer) i.next ();
          intVal = integerVal.intValue ();
          p = (Path) _pi4 [intVal].getFirst ();
          temp = (Path) _pi4 [v].getFirst ();

          if (p != temp.getBack ()) {
            _c4 [v].add (p);
          }
        }
      }

      temp = (Path) _pi4 [v].get (k - 1);
      p = temp.getBack ();

      if (p != null) {
        intVal = p.getNode ();
        kp = _pi4 [intVal].indexOf (p);

        if ((kp + 1) == _pi4 [intVal].size ()) {
          nextPath4 (intVal, kp + 1);
        }

        temp = (Path) _pi4 [intVal].get (kp + 1);

        if (temp.getBack () != null) {
          _c4 [v].add (temp);
        }
      }
    }

    if (_c4 [v].size () != 0) {
      Path min = (Path) _c4 [v].getFirst ();
      float l;

      for (ListIterator i = _c4 [v].listIterator (); i.hasNext (); ) {
        temp = (Path) i.next ();

        if (temp.getLenght () < min.getLenght ()) {
          min = temp;
        }
      }

      l = min.getLenght ();
      n = (Node) this.getGraph ().getNodeByPos (v);
      l += n.getWeight ();
      temp = new Path (min, l, v);

      _pi4 [v].add (temp);
      _c4 [v].remove (min);
    } else {
      temp = new Path (v);

      _pi4 [v].add (temp);
    }
  }


  /**
   * nextPath3 Method
   * @param v the node.
   * @param k the current position.
   */
  public void nextPath3 (int v, int k) {

    Integer integerVal;
    Node n;
    int intVal, kp;
    Path p, temp;

    if (v != this._source3) {
      if (k == 1) {
        n = (Node) this.getGraph ().getNodeByPos (v);
        this._c3 [v] = new LinkedList ();

        for (ListIterator i = n.getPrev ().listIterator (); i.hasNext (); ) {
          integerVal = (Integer) i.next ();
          intVal = integerVal.intValue ();
          p = (Path) _pi3 [intVal].getFirst ();
          temp = (Path) _pi3 [v].getFirst ();

          if (p != temp.getBack ()) {
            _c3 [v].add (p);
          }
        }
      }

      temp = (Path) _pi3 [v].get (k - 1);
      p = temp.getBack ();

      if (p != null) {
        intVal = p.getNode ();
        kp = _pi3 [intVal].indexOf (p);

        if ((kp + 1) == _pi3 [intVal].size ()) {
          nextPath3 (intVal, kp + 1);
        }

        temp = (Path) _pi3 [intVal].get (kp + 1);

        if (temp.getBack () != null) {
          _c3 [v].add (temp);
        }
      }
    }

    if (_c3 [v].size () != 0) {
      Path min = (Path) _c3 [v].getFirst ();
      float l;

      for (ListIterator i = _c3 [v].listIterator (); i.hasNext (); ) {
        temp = (Path) i.next ();

        if (temp.getLenght () < min.getLenght ()) {
          min = temp;
        }
      }

      l = min.getLenght ();
      n = (Node) this.getGraph ().getNodeByPos (v);
      l += n.getWeight ();
      temp = new Path (min, l, v);

      _pi3 [v].add (temp);
      _c3 [v].remove (min);
    } else {
      temp = new Path (v);

      _pi3 [v].add (temp);
    }
  }


  /**
   * nextPath2 Method
   * @param v the node.
   * @param k the current position.
   */
  public void nextPath2 (int v, int k) {

    Integer integerVal;
    Node n;
    int intVal, kp;
    Path p, temp;

    if (v != this._source2) {
      if (k == 1) {
        n = (Node) this.getGraph ().getNodeByPos (v);
        this._c2 [v] = new LinkedList ();

        for (ListIterator i = n.getPrev ().listIterator (); i.hasNext (); ) {
          integerVal = (Integer) i.next ();
          intVal = integerVal.intValue ();
          p = (Path) _pi2 [intVal].getFirst ();
          temp = (Path) _pi2 [v].getFirst ();

          if (p != temp.getBack ()) {
            _c2 [v].add (p);
          }
        }
      }

      temp = (Path) _pi2 [v].get (k - 1);
      p = temp.getBack ();

      if (p != null) {
        intVal = p.getNode ();
        kp = _pi2 [intVal].indexOf (p);

        if ((kp + 1) == _pi2 [intVal].size ()) {
          nextPath2 (intVal, kp + 1);
        }

        temp = (Path) _pi2 [intVal].get (kp + 1);

        if (temp.getBack () != null) {
          _c2 [v].add (temp);
        }
      }
    }

    if (_c2 [v].size () != 0) {
      Path min = (Path) _c2 [v].getFirst ();
      float l;

      for (ListIterator i = _c2 [v].listIterator (); i.hasNext (); ) {
        temp = (Path) i.next ();

        if (temp.getLenght () < min.getLenght ()) {
          min = temp;
        }
      }

      l = min.getLenght ();
      n = (Node) this.getGraph ().getNodeByPos (v);
      l += n.getWeight ();
      temp = new Path (min, l, v);

      _pi2 [v].add (temp);
      _c2 [v].remove (min);
    } else {
      temp = new Path (v);

      _pi2 [v].add (temp);
    }
  }





  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////
  // 
  // Needed structures for REA method
  // 
  
  /**  */
  private boolean _isInitialized = false;

  /** The current graph */
  private Graph _graph;            // to initialize

  /** The source node id */
  private int _source;             // to initialize

  /** The start node id */
  private String _start;           // to initialize

  /** The end node id */
  private String _end;             // to initialize

  /** The number of solution to find */
  private int _nbSolutions;        // to initialize
  private boolean _byRank = false; 

  /** Field C[] */
  private LinkedList _c [];

  /** Field PI[] */
  private LinkedList _pi [];

  /** Field timeout */
  private long _timeout = 30000000;  // to remove

  /** Field timeOver */
  private boolean _timeOver;       // to remove

  // 
  // Needed Structures for verifyConstraints method.
  // 

  /** Field limit */
  private float _limit = 1000;     // to initialize

  /** Field levelMax */
  private int _levelMax = 9;       // to initialize

  // 
  // Needed Structures for builtpath method.
  // 

  /** Field currentWay */
  private LinkedList _currentWay;

  /** Field currentDistance */
  private float _currentDistance;

  // 
  // Needed Structures for builtpath method.
  // 

  /** Field solutions */
  private ArrayList _solutions;    // to initialize
  private TreeMap _tempsolutionsmap;
  private TreeMap _solutionsMap;
  //
  // Needed structures for research between reactions.
  //
  
  /** Field C2[] */
  private LinkedList _c2 [];

  /** Field PI2[] */
  private LinkedList _pi2 [];

  /** Field C3[] */
  private LinkedList _c3 [];

  /** Field PI3[] */
  private LinkedList _pi3 [];

  /** Field C4[] */
  private LinkedList _c4 [];

  /** Field PI4[] */
  private LinkedList _pi4 [];
  
  /** Field source2 */
  private int _source2; // to initialize

  /** Field source3 */
  private int _source3; // to initialize

  /** Field source4 */
  private int _source4; // to initialize
  
  private boolean finnished = false;
/**
 * @return Returns the timeout.
 */
public long getTimeout() {
	return _timeout;
}
/**
 * @param timeout The timeout to set.
 */
public void setTimeout(long timeout) {
	_timeout = timeout;
}
/**
 * @return Returns the _byRank.
 */
public boolean isByRank() {
	return _byRank;
}
/**
 * @param rank The _byRank to set.
 */
public void setByRank(boolean byrank) {
	_byRank = byrank;
}
/**
 * @return Returns the _solutionsMap.
 */
public TreeMap getSolutionsMap() {
	return _solutionsMap;
}
/**
 * @param map The _solutionsMap to set.
 */
public void setSolutionsMap(TreeMap map) {
	_solutionsMap = map;
}

/* (non-Javadoc)
 * @see java.lang.Runnable#run()
 */
public void run() {
	try {
		// TODO Auto-generated method stub
		this.call();
	} catch (Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
}
}


/*--- Formatted in North Style ---*/
