/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core;



import java.io.ByteArrayOutputStream;
import java.io.IOException;

import java.util.HashMap;

import net.sourceforge.gxl.GXL;

import org.apache.crimson.tree.DOMImplementationImpl;
import org.apache.crimson.tree.XmlDocument;

import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Text;


/**
 * Class of support to export an IGraph in different formats.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: ExportSupport.java,v 1.13 2004/11/08 14:40:03 frederic Exp $
 */
public class ExportSupport {

  /**
   * Returns the graph in the AT&T DOT format.
   * @param graph the graph to be exported.
   * @return the graph in the AT&T DOT format in a string.
   */
  public static String toDot (IExportableSubdividedGraph graph) {

    Document document = new XmlDocument ();
    Element graphElt = graph.toElement (document);
    StringBuffer buffer = new StringBuffer ();
    StringBuffer graphAttrBuffer = new StringBuffer ();
    StringBuffer nodesBuffer = new StringBuffer ();
    StringBuffer arcsBuffer = new StringBuffer ();
    StringBuffer subgraphsBuffer = new StringBuffer ();

    // begin digraph
    buffer.append ("digraph \"");
    buffer.append (graphElt.getAttribute (IDOM.ID));
    buffer.append ("\" {\n");

    // Has TYPE ?
    if (graphElt.hasAttribute (IDOM.TYPE)) {
      graphAttrBuffer.append ("  \"");
      graphAttrBuffer.append (IDOM.TYPE);
      graphAttrBuffer.append ("\" = \"");
      graphAttrBuffer.append (graphElt.getAttribute (IDOM.TYPE));
      graphAttrBuffer.append ("\"\n");
    }

    Element element = (Element) graphElt.getFirstChild ();

    while (element != null) {
      String tagName = element.getTagName ();

      if (tagName == IDOM.ATTR) {

        // "attributeName" = "value"
        graphAttrBuffer.append ("  ");
        ExportSupport.dotProcessAttrElement (graphAttrBuffer, element);
        graphAttrBuffer.append ('\n');
      } else if (tagName == IDOM.ATTR_LIST) {

        // "attributeName" = "value|value"
        graphAttrBuffer.append ("  ");
        ExportSupport.dotProcessAttrListElement (graphAttrBuffer, element);
        graphAttrBuffer.append ('\n');
      } else if (tagName == IDOM.NODE) {
        ExportSupport.dotProcessNodeElement (nodesBuffer, element);
      } else if (tagName == IDOM.ARC) {
        ExportSupport.dotProcessArcElement (arcsBuffer, element);
      } else if (tagName == IDOM.SUBGRAPH) {
        ExportSupport.dotProcessSubgraphElement (subgraphsBuffer, element);
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      element = (Element) element.getNextSibling ();
    }

    buffer.append (graphAttrBuffer);
    buffer.append (nodesBuffer);
    buffer.append (arcsBuffer);
    buffer.append (subgraphsBuffer);

    // end digraph
    buffer.append ("}");

    String result = buffer.toString ();

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Process the 'attr' DOM element and append result to the given buffer for the DOT export.
   * @param buffer the working buffer.
   * @param attrElt the 'attr' DOM element
   */
  protected static void dotProcessAttrElement (StringBuffer buffer, Element attrElt) {

    // 
    assert buffer != null;

    assert attrElt != null;

    assert attrElt.getTagName () == IDOM.ATTR;

    // 
    String name = attrElt.getAttribute (IDOM.NAME);

    if (name.equalsIgnoreCase (ICharacterized.LABEL_KEY)) {
      name = ICharacterized.LABEL_KEY;
    }

    buffer.append ('\"');
    buffer.append (name);
    buffer.append ("\" = \"");
    buffer.append (attrElt.getAttribute (IDOM.VALUE));
    buffer.append ('\"');
  }

  /**
   * Process the 'attr-list' DOM element and append result to the given buffer for the DOT export.
   * @param buffer the working buffer.
   * @param attrListElt the 'attr-list' DOM element
   */
  protected static void dotProcessAttrListElement (StringBuffer buffer, Element attrListElt) {

    // 
    assert buffer != null;

    assert attrListElt != null;

    assert attrListElt.getTagName () == IDOM.ATTR_LIST;

    // 
    buffer.append ('\"');
    buffer.append (attrListElt.getAttribute (IDOM.NAME));
    buffer.append ("\" = \"");

    boolean flagPipe = false;
    Element valueElt = (Element) attrListElt.getFirstChild ();

    while (valueElt != null) {
      if (flagPipe) {
        buffer.append ('|');
      } else {
        flagPipe = true;
      }

      buffer.append (valueElt.getFirstChild ().getNodeValue ());

      valueElt = (Element) valueElt.getNextSibling ();
    }

    buffer.append ('\"');
  }

  /**
   * Process the 'node' DOM element and append result to the given buffer for the DOT export.
   * @param buffer the working buffer.
   * @param nodeElt the 'node' DOM element
   */
  protected static void dotProcessNodeElement (StringBuffer buffer, Element nodeElt) {

    // 
    assert buffer != null;

    assert nodeElt != null;

    assert nodeElt.getTagName () == IDOM.NODE;

    // 
    // Do not record nodes without attributes
    buffer.append ("  \"");
    buffer.append (nodeElt.getAttribute (IDOM.ID));
    buffer.append ("\"");
    ExportSupport.dotProcessAttributes (buffer, nodeElt);
    buffer.append ("\n");
  }

  /**
   * Process the 'arc' DOM element and append result to the given buffer for the DOT export.
   * @param buffer the working buffer.
   * @param arcElt the 'arc' DOM element
   */
  protected static void dotProcessArcElement (StringBuffer buffer, Element arcElt) {

    // 
    assert buffer != null;

    assert arcElt != null;

    assert arcElt.getTagName () == IDOM.ARC;

    // 
    buffer.append ("  \"");
    buffer.append (arcElt.getAttribute (IDOM.TAIL));
    buffer.append ("\" -> \"");
    buffer.append (arcElt.getAttribute (IDOM.HEAD));
    buffer.append ("\"");
    ExportSupport.dotProcessAttributes (buffer, arcElt);
    buffer.append ("\n");
  }

  /**
   * Process the 'attr' and 'attr-list' child elements from 'arc' and 'node' elements, and the 'type' and 'stamp' attributes as well.
   * @param buffer the working buffer.
   * @param containerElt either an 'arc' element or a 'node' element.
   */
  protected static void dotProcessAttributes (StringBuffer buffer, Element containerElt) {

    // 
    assert buffer != null;

    assert containerElt != null;

    assert (containerElt.getTagName () == IDOM.NODE) || (containerElt.getTagName () == IDOM.ARC);

    // 
    StringBuffer attrBuffer = new StringBuffer ();

    attrBuffer.append ("[");

    boolean flagComma = false;
    Element attrElt = (Element) containerElt.getFirstChild ();

    // Has type ?
    if (containerElt.hasAttribute (IDOM.TYPE)) {
      attrBuffer.append ('\"');
      attrBuffer.append (IDOM.TYPE);
      attrBuffer.append ("\" = \"");
      attrBuffer.append (containerElt.getAttribute (IDOM.TYPE));
      attrBuffer.append ('\"');

      flagComma = true;
    }

    // Has stamp ?
    if (containerElt.hasAttribute (IDOM.STAMP)) {
      if (flagComma) {
        attrBuffer.append (", ");
      }

      attrBuffer.append ('\"');
      attrBuffer.append (IDOM.STAMP);
      attrBuffer.append ("\" = \"");
      attrBuffer.append (containerElt.getAttribute (IDOM.STAMP));
      attrBuffer.append ('\"');

      flagComma = true;
    }

    while (attrElt != null) {
      String tagName = attrElt.getTagName ();

      if (tagName == IDOM.ATTR) {
        if (flagComma) {
          attrBuffer.append (", ");
        } else {
          flagComma = true;
        }

        // "attributeName" = "value"
        ExportSupport.dotProcessAttrElement (attrBuffer, attrElt);
      } else if (tagName == IDOM.ATTR_LIST) {
        if (flagComma) {
          attrBuffer.append (", ");
        } else {
          flagComma = true;
        }

        // "attributeName" = "value, value"
        ExportSupport.dotProcessAttrListElement (attrBuffer, attrElt);
      } else if (tagName == IDOM.REFERENCE) {

        // Do nothing
      } else if (tagName == IDOM.REFERENCE_SET) {

        // Do nothing
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      attrElt = (Element) attrElt.getNextSibling ();
    }

    attrBuffer.append ("]");

    // If there's at least one attribute, the flagComma is set.
    if (flagComma) {
      buffer.append (" ");
      buffer.append (attrBuffer);
    }
  }

  /**
   * Process the 'subgraph' DOM element and append result to the given buffer for the DOT export.
   * @param buffer the working buffer.
   * @param subgraphElt the 'subgraph' DOM element
   */
  protected static void dotProcessSubgraphElement (StringBuffer buffer, Element subgraphElt) {

    // 
    assert buffer != null;

    assert subgraphElt != null;

    assert subgraphElt.getTagName () == IDOM.SUBGRAPH;

    // 
    // buffers
    StringBuffer subgraphAttrBuffer = new StringBuffer ();
    StringBuffer sgnodesBuffer = new StringBuffer ();
    StringBuffer sgarcsBuffer = new StringBuffer ();

    // begin digraph
    buffer.append ("  subgraph ");
    buffer.append (subgraphElt.getAttribute (IDOM.ID));
    buffer.append (" {\n");

    // Has TYPE ?
    if (subgraphElt.hasAttribute (IDOM.TYPE)) {
      subgraphAttrBuffer.append ("    \"");
      subgraphAttrBuffer.append (IDOM.TYPE);
      subgraphAttrBuffer.append ("\" = \"");
      subgraphAttrBuffer.append (subgraphElt.getAttribute (IDOM.TYPE));
      subgraphAttrBuffer.append ("\"\n");
    }

    Element element = (Element) subgraphElt.getFirstChild ();

    while (element != null) {
      String tagName = element.getTagName ();

      if (tagName == IDOM.ATTR) {

        // "attributeName" = "value"
        subgraphAttrBuffer.append ("    ");
        ExportSupport.dotProcessAttrElement (subgraphAttrBuffer, element);
        subgraphAttrBuffer.append ('\n');
      } else if (tagName == IDOM.ATTR_LIST) {

        // "attributeName" = "value, value"
        subgraphAttrBuffer.append ("    ");
        ExportSupport.dotProcessAttrListElement (subgraphAttrBuffer, element);
        subgraphAttrBuffer.append ('\n');
      } else if (tagName == IDOM.SG_NODE) {
        sgnodesBuffer.append ("    \"");
        sgnodesBuffer.append (element.getAttribute (IDOM.REF));
        sgnodesBuffer.append ("\"\n");
      } else if (tagName == IDOM.SG_ARC) {

        // Do nothing
        // because, unfornately, DOT is limited, and duplicates arcs from subgraphs.
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      element = (Element) element.getNextSibling ();
    }

    buffer.append (subgraphAttrBuffer);
    buffer.append (sgnodesBuffer);
    buffer.append (sgarcsBuffer);

    // end subgraph
    buffer.append ("  }\n");
  }

  /**
   * Returns the graph in the GML format.
   * @param graph the graph to be exported.
   * @return the graph in the GML format in a string.
   */
  public static String toGML (IExportableSubdividedGraph graph) {

    Document document = new XmlDocument ();
    Element graphElt = graph.toElement (document);
    StringBuffer buffer = new StringBuffer ();
    StringBuffer graphAttrBuffer = new StringBuffer ();
    StringBuffer nodesBuffer = new StringBuffer ();
    StringBuffer arcsBuffer = new StringBuffer ();

    // begin graph
    buffer.append ("graph [\n");
    buffer.append ("  directed 1\n");

    // Has TYPE ?
    if (graphElt.hasAttribute (IDOM.TYPE)) {

      // type typeName
      graphAttrBuffer.append ("  ");
      graphAttrBuffer.append (IDOM.TYPE);
      graphAttrBuffer.append (" \"");
      graphAttrBuffer.append (graphElt.getAttribute (IDOM.TYPE));
      graphAttrBuffer.append ("\"\n");
    }

    // Map containing nodes' stringId vs. intId
    HashMap nodeIds = new HashMap ();
    int nodeNum = 1;
    Element element = (Element) graphElt.getFirstChild ();

    while (element != null) {
      String tagName = element.getTagName ();

      if (tagName == IDOM.ATTR) {

        // attributeName value
        graphAttrBuffer.append ("  ");
        ExportSupport.gmlProcessAttrElement (graphAttrBuffer, element);
        graphAttrBuffer.append ('\n');
      } else if (tagName == IDOM.ATTR_LIST) {

        // attributeName [
        // element [value]
        // element [value]
        // ]
        graphAttrBuffer.append ("  ");
        ExportSupport.gmlProcessAttrListElement (graphAttrBuffer, element);
      } else if (tagName == IDOM.NODE) {
        String nodeId = element.getAttribute (IDOM.ID);

        nodeIds.put (nodeId, new Integer (nodeNum));
        ExportSupport.gmlProcessNodeElement (nodesBuffer, element, nodeNum);

        ++nodeNum;
      } else if (tagName == IDOM.ARC) {
        ExportSupport.gmlProcessArcElement (arcsBuffer, element, nodeIds);
      } else if (tagName == IDOM.SUBGRAPH) {

        // Do nothing: Not supported by GML.
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      element = (Element) element.getNextSibling ();
    }

    buffer.append (graphAttrBuffer);
    buffer.append (nodesBuffer);
    buffer.append (arcsBuffer);

    // end graph
    buffer.append ("]\n");

    String result = buffer.toString ();

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Process the 'attr' DOM element and append result to the given buffer for the GML export.
   * @param buffer the working buffer.
   * @param attrElt the 'attr' DOM element
   */
  protected static void gmlProcessAttrElement (StringBuffer buffer, Element attrElt) {

    // 
    assert buffer != null;

    assert attrElt != null;

    assert attrElt.getTagName () == IDOM.ATTR;

    // 
    String name = attrElt.getAttribute (IDOM.NAME);

    if (name.equalsIgnoreCase (ICharacterized.LABEL_KEY)) {
      name = ICharacterized.LABEL_KEY;
    }

    buffer.append (name);
    buffer.append (" \"");
    buffer.append (attrElt.getAttribute (IDOM.VALUE));
    buffer.append ("\"\n");
  }

  /**
   * Process the 'attr-list' DOM element and append result to the given buffer for the GML export.
   * @param buffer the working buffer.
   * @param attrListElt the 'attr-list' DOM element
   */
  protected static void gmlProcessAttrListElement (StringBuffer buffer, Element attrListElt) {

    // 
    assert buffer != null;

    assert attrListElt != null;

    assert attrListElt.getTagName () == IDOM.ATTR_LIST;

    // attributeName [
    // element [value]
    // element [value]
    // ]
    buffer.append (attrListElt.getAttribute (IDOM.NAME));
    buffer.append (" [\n");

    Element valueElt = (Element) attrListElt.getFirstChild ();

    while (valueElt != null) {
      buffer.append ("    item [\"");
      buffer.append (valueElt.getFirstChild ().getNodeValue ());
      buffer.append ("\"]\n");

      valueElt = (Element) valueElt.getNextSibling ();
    }

    buffer.append ("  ]\n");
  }

  /**
   * Process the 'node' DOM element and append result to the given buffer for the GML export.
   * @param buffer the working buffer.
   * @param nodeElt the 'node' DOM element.
   * @param id the integer id of the node.
   */
  protected static void gmlProcessNodeElement (StringBuffer buffer, Element nodeElt, int id) {

    // 
    assert buffer != null;

    assert nodeElt != null;

    assert nodeElt.getTagName () == IDOM.NODE;

    // 
    buffer.append ("  node [\n    id ");
    buffer.append (id);
    buffer.append ("\n");
    ExportSupport.gmlProcessAttributes (buffer, nodeElt);
    buffer.append ("  ]\n");
  }

  /**
   * Process the 'arc' DOM element and append result to the given buffer for the GML export.
   * @param buffer the working buffer.
   * @param arcElt the 'arc' DOM element
   * @param map the map String nodeId vs. int nodeId
   */
  protected static void gmlProcessArcElement (StringBuffer buffer, Element arcElt, HashMap map) {

    // 
    assert buffer != null;

    assert arcElt != null;

    assert map != null;

    assert arcElt.getTagName () == IDOM.ARC;

    // 
    buffer.append ("  edge [\n    source ");
    buffer.append (map.get (arcElt.getAttribute (IDOM.TAIL)));
    buffer.append ("\n    target ");
    buffer.append (map.get (arcElt.getAttribute (IDOM.HEAD)));
    buffer.append ("\n");
    ExportSupport.gmlProcessAttributes (buffer, arcElt);
    buffer.append ("  ]\n");
  }

  /**
   * Process the 'attr' and 'attr-list' child elements from 'arc' and 'node' elements, and the 'type' and 'stamp' attributes as well.
   * @param buffer the working buffer.
   * @param containerElt either an 'arc' element or a 'node' element.
   */
  protected static void gmlProcessAttributes (StringBuffer buffer, Element containerElt) {

    // 
    assert buffer != null;

    assert containerElt != null;

    assert (containerElt.getTagName () == IDOM.NODE) || (containerElt.getTagName () == IDOM.ARC);

    // 
    // Alternate id
    if (containerElt.getTagName () == IDOM.NODE) {
      buffer.append ("    alternateId \"");
      buffer.append (containerElt.getAttribute (IDOM.ID));
      buffer.append ("\"\n");

      // Has stamp ?
      if (containerElt.hasAttribute (IDOM.STAMP)) {
        buffer.append ("    ");
        buffer.append (IDOM.STAMP);
        buffer.append (" \"");
        buffer.append (containerElt.getAttribute (IDOM.STAMP));
        buffer.append ("\"\n");
      }
    }

    // Has type ?
    if (containerElt.hasAttribute (IDOM.TYPE)) {
      buffer.append ("    ");
      buffer.append (IDOM.TYPE);
      buffer.append (" \"");
      buffer.append (containerElt.getAttribute (IDOM.TYPE));
      buffer.append ("\"\n");
    }

    Element attrElt = (Element) containerElt.getFirstChild ();

    while (attrElt != null) {
      String tagName = attrElt.getTagName ();

      if (tagName == IDOM.ATTR) {

        // attributeName value
        buffer.append ("    ");
        ExportSupport.gmlProcessAttrElement (buffer, attrElt);
      } else if (tagName == IDOM.ATTR_LIST) {

        // attributeName [
        // element [value]
        // element [value]
        // ]
        buffer.append ("    ");
        ExportSupport.gmlProcessAttrListElement (buffer, attrElt);
      } else if (tagName == IDOM.REFERENCE) {

        // Do nothing
      } else if (tagName == IDOM.REFERENCE_SET) {

        // Do nothing
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      attrElt = (Element) attrElt.getNextSibling ();
    }
  }

  /**
   * Returns the graph in the GXL format.
   * @param graph the graph to be exported.
   * @return the graph in the GXL format in a string.
   */
  public static String toGXL (IExportableSubdividedGraph graph) {

    // 
    assert graph != null;

    // 
    DOMImplementation domImpl = DOMImplementationImpl.getDOMImplementation ();
    DocumentType docType = domImpl.createDocumentType ("gxl", null, "http://www.gupro.de/GXL/gxl-1.0.dtd");
    Document gxl = domImpl.createDocument (null, "gxl", docType);
    Element gxlElt = gxl.getDocumentElement ();

    gxlElt.setAttribute ("xmlns:xlink", "http://www.w3.org/1999/xlink");

    Document irs = new XmlDocument ();
    Element graphElt = graph.toElement (irs);
    Element gxlGraphElt = gxl.createElement (GXL.GRAPH);

    gxlGraphElt.setAttribute (GXL.ID, graphElt.getAttribute (IDOM.ID));
    gxlGraphElt.setAttribute (GXL.EDGEIDS, GXL.TRUE);
    gxlElt.appendChild (gxlGraphElt);

    Element element = (Element) graphElt.getFirstChild ();

    while (element != null) {
      String tagName = element.getTagName ();

      if (tagName == IDOM.ATTR) {

        // graph attribute
        ExportSupport.gxlProcessAttrElement (gxl, gxlGraphElt, element);
      } else if (tagName == IDOM.ATTR_LIST) {

        // graph multi-attribute
        ExportSupport.gxlProcessAttrListElement (gxl, gxlGraphElt, element);
      } else if (tagName == IDOM.NODE) {
        ExportSupport.gxlProcessNodeElement (gxl, gxlGraphElt, element);
      } else if (tagName == IDOM.ARC) {
        ExportSupport.gxlProcessArcElement (gxl, gxlGraphElt, element);
      } else if (tagName == IDOM.SUBGRAPH) {

        // Do nothing: Not supported by GXL.
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      element = (Element) element.getNextSibling ();
    }

    ByteArrayOutputStream baos = new ByteArrayOutputStream ();

    try {
      ((XmlDocument) gxl).write (baos);
    } catch (IOException e) {
      throw new RuntimeException ("In-Memory serialization failed", e);
    }

    String result = baos.toString ();

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Process the 'attr' DOM element and append result to the given buffer for the GML export.
   * @param document the parent document.
   * @param parentElt the parent element.
   * @param attrElt the 'attr' DOM element
   */
  protected static void gxlProcessAttrElement (Document document, Element parentElt, Element attrElt) {

    // 
    assert document != null;

    assert parentElt != null;

    assert attrElt != null;

    assert attrElt.getTagName () == IDOM.ATTR;

    // 
    // <attr name="attrName">
    // <string>value</string>
    // </attr>
    Element element = document.createElement (GXL.ATTR);

    element.setAttribute (GXL.NAME, attrElt.getAttribute (IDOM.NAME));

    Element string = document.createElement (GXL.STRING);
    Text value = document.createTextNode (attrElt.getAttribute (IDOM.VALUE));

    string.appendChild (value);
    element.appendChild (string);
    parentElt.appendChild (element);
  }

  /**
   * Process the 'attr-list' DOM element and append result to the given buffer for the GML export.
   * @param document the parent document.
   * @param parentElt the parent element.
   * @param attrListElt the 'attr-list' DOM element
   */
  protected static void gxlProcessAttrListElement (Document document, Element parentElt, Element attrListElt) {

    // 
    assert document != null;

    assert parentElt != null;

    assert attrListElt != null;

    assert attrListElt.getTagName () == IDOM.ATTR_LIST;

    // 
    // <attr name="attrName">
    // <seq>
    // <string>value1</string>
    // <string>value2</string>
    // <string>value3</string>
    // <string>value4</string>
    // </seq>
    // </attr>
    Element attr = document.createElement (GXL.ATTR);

    attr.setAttribute (GXL.NAME, attrListElt.getAttribute (IDOM.NAME));

    Element seq = document.createElement (GXL.SEQ);
    Element valueElt = (Element) attrListElt.getFirstChild ();

    while (valueElt != null) {
      Element string = document.createElement (GXL.STRING);
      Text value = document.createTextNode (valueElt.getFirstChild ().getNodeValue ());

      string.appendChild (value);
      seq.appendChild (string);

      valueElt = (Element) valueElt.getNextSibling ();
    }

    attr.appendChild (seq);
    parentElt.appendChild (attr);
  }

  /**
   * Process the 'node' DOM element and append result to the given buffer for the GML export.
   * @param document the parent document.
   * @param parentElt the parent element.
   * @param nodeElt the 'node' DOM element.
   */
  protected static void gxlProcessNodeElement (Document document, Element parentElt, Element nodeElt) {

    // 
    assert document != null;

    assert parentElt != null;

    assert nodeElt != null;

    assert nodeElt.getTagName () == IDOM.NODE;

    // 
    // <node id="nodeId">
    // <attr name="attrName">
    // <string>value</string>
    // </attr>
    // <attr name="multiAttrName">
    // <seq>
    // <string>value1</string>
    // <string>value2</string>
    // <string>value3</string>
    // <string>value4</string>
    // </seq>
    // </attr>
    // </node>
    Element node = document.createElement (GXL.NODE);

    node.setAttribute (GXL.ID, nodeElt.getAttribute (IDOM.ID));
    ExportSupport.gxlProcessAttributes (document, node, nodeElt);
    parentElt.appendChild (node);
  }

  /**
   * Process the 'arc' DOM element and append result to the given buffer for the GML export.
   * @param document the parent document.
   * @param parentElt the parent element.
   * @param arcElt the 'arc' DOM element.
   */
  protected static void gxlProcessArcElement (Document document, Element parentElt, Element arcElt) {

    // 
    assert document != null;

    assert parentElt != null;

    assert arcElt != null;

    assert arcElt.getTagName () == IDOM.ARC;

    // 
    Element edge = document.createElement (GXL.EDGE);

    edge.setAttribute (GXL.ID, arcElt.getAttribute (IDOM.ID));
    edge.setAttribute (GXL.FROM, arcElt.getAttribute (IDOM.TAIL));
    edge.setAttribute (GXL.TO, arcElt.getAttribute (IDOM.HEAD));
    ExportSupport.gxlProcessAttributes (document, edge, arcElt);
    parentElt.appendChild (edge);
  }

  /**
   * Process the 'attr' and 'attr-list' child elements from 'arc' and 'node' elements, and the 'type' and 'stamp' attributes as well.
   * @param document the parent document.
   * @param parentElt the parent element.
   * @param containerElt either an 'arc' element or a 'node' element.
   */
  protected static void gxlProcessAttributes (Document document, Element parentElt, Element containerElt) {

    // 
    assert document != null;

    assert parentElt != null;

    assert containerElt != null;

    assert (parentElt.getTagName () == GXL.NODE) || (parentElt.getTagName () == GXL.EDGE);

    assert (containerElt.getTagName () == IDOM.NODE) || (containerElt.getTagName () == IDOM.ARC);

    // Has stamp ?
    if (containerElt.hasAttribute (IDOM.STAMP)) {
      Element stampAttr = document.createElement (GXL.ATTR);

      stampAttr.setAttribute (GXL.NAME, IDOM.STAMP);

      Element string = document.createElement (GXL.STRING);
      Text value = document.createTextNode (containerElt.getAttribute (IDOM.STAMP));

      string.appendChild (value);
      stampAttr.appendChild (string);
      parentElt.appendChild (stampAttr);
    }

    // Has type ?
    if (containerElt.hasAttribute (IDOM.TYPE)) {
      Element typeAttr = document.createElement (GXL.ATTR);

      typeAttr.setAttribute (GXL.NAME, IDOM.TYPE);

      Element string = document.createElement (GXL.STRING);
      Text value = document.createTextNode (containerElt.getAttribute (IDOM.TYPE));

      string.appendChild (value);
      typeAttr.appendChild (string);
      parentElt.appendChild (typeAttr);
    }

    Element attrElt = (Element) containerElt.getFirstChild ();

    while (attrElt != null) {
      String tagName = attrElt.getTagName ();

      if (tagName == IDOM.ATTR) {
        ExportSupport.gxlProcessAttrElement (document, parentElt, attrElt);
      } else if (tagName == IDOM.ATTR_LIST) {
        ExportSupport.gxlProcessAttrListElement (document, parentElt, attrElt);
      } else if (tagName == IDOM.REFERENCE) {

        // Do nothing
      } else if (tagName == IDOM.REFERENCE_SET) {

        // Do nothing
      } else {
        throw new IllegalArgumentException ("Wrong graph!");
      }

      attrElt = (Element) attrElt.getNextSibling ();
    }
  }
}


/*--- Formatted in North Style ---*/
