/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core;



import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;

import java.util.Iterator;

import north.util.Map;


/**
 * Repository for all kind of graphs.
 * Indexed by their ids
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: GraphRepository.java,v 1.4 2004/12/09 13:28:17 fabian Exp $
 */
public class GraphRepository
  extends Map
  implements PropertyChangeListener {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Constructor.
   */
  protected GraphRepository () {
  }

  /**
   * Creates a new instance.
   * @return the new instance.
   */
  public static GraphRepository newGraphRepository () {

    GraphRepository result = new GraphRepository ();

    result.init (String.class, IGraph.class);

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Returns the singleton instance.
   * @return the singleton instance.
   */
  public static GraphRepository getInstance () {
    return GraphRepository.INSTANCE;
  }

  /**
   * Returns the propertyChangeSupport.
   * @return the propertyChangeSupport.
   */
  protected PropertyChangeSupport getPropertyChangeSupport () {

    // 
    assert this._propertyChangeSupport != null;

    // 
    return this._propertyChangeSupport;
  }

  /**
   * Returns the vetoableChangeSupport.
   * @return the vetoableChangeSupport.
   */
  protected VetoableChangeSupport getVetoableChangeSupport () {

    // 
    assert this._vetoableChangeSupport != null;

    // 
    return this._vetoableChangeSupport;
  }

  // //////////////////////////////////////////////////////////
  // 
  // JavaBeans
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Adds a PropertyChangeListener on the GraphRepository.
   * For an easy implementation @see java.beans.PropertyChangeSupport#addPropertyChangeListener(java.beans.PropertyChangeListener)
   * @param listener the listener to be added.
   */
  public void addPropertyChangeListener (PropertyChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getPropertyChangeSupport ().addPropertyChangeListener (listener);
  }

  /**
   * Adds a PropertyChangeListener on the GraphRepository.
   * For an easy implementation @see java.beans.PropertyChangeSupport#addPropertyChangeListener(java.lang.String, java.beans.PropertyChangeListener)
   * @param propertyName the property name.
   * @param listener the listener to be added.
   */
  public void addPropertyChangeListener (String propertyName, PropertyChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getPropertyChangeSupport ().addPropertyChangeListener (propertyName, listener);
  }

  /**
   * Removes a PropertyChangeListener from the GraphRepository.
   * For an easy implementation @see java.beans.PropertyChangeSupport#removePropertyChangeListener(java.beans.PropertyChangeListener)
   * @param listener the listener to be removed.
   */
  public void removePropertyChangeListener (PropertyChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getPropertyChangeSupport ().removePropertyChangeListener (listener);
  }

  /**
   * Removes a PropertyChangeListener from the GraphRepository.
   * For an easy implementation @see java.beans.PropertyChangeSupport#removePropertyChangeListener(java.lang.String, java.beans.PropertyChangeListener)
   * @param propertyName the property name.
   * @param listener the listener to be removed.
   */
  public void removePropertyChangeListener (String propertyName, PropertyChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getPropertyChangeSupport ().removePropertyChangeListener (propertyName, listener);
  }

  /**
   * Reports a bound property update to any registered listeners.
   * For an easy implementation @see java.beans.PropertyChangeSupport#firePropertyChange(java.lang.String, java.lang.Object, java.lang.Object)
   * @param propertyName the property name.
   * @param oldValue the old value.
   * @param newValue the new value.
   */
  public void firePropertyChange (String propertyName, Object oldValue, Object newValue) {

    // 
    assert propertyName != null;

    // 
    this.getPropertyChangeSupport ().firePropertyChange (propertyName, oldValue, newValue);
  }

  /**
   * Adds a VetoableChangeListener on the GraphRepository.
   * For an easy implementation @see java.beans.VetoableChangeSupport#addVetoableChangeListener(java.beans.VetoableChangeListener)
   * @param listener the listener to be added.
   */
  public void addVetoableChangeListener (VetoableChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getVetoableChangeSupport ().addVetoableChangeListener (listener);
  }

  /**
   * Adds a VetoableChangeListener on the GraphRepository.
   * For an easy implementation @see java.beans.VetoableChangeSupport#addVetoableChangeListener(java.lang.String, java.beans.VetoableChangeListener)
   * @param propertyName the property name.
   * @param listener the listener to be added.
   */
  public void addVetoableChangeListener (String propertyName, VetoableChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getVetoableChangeSupport ().addVetoableChangeListener (propertyName, listener);
  }

  /**
   * Removes a VetoableChangeListener from the GraphRepository.
   * For an easy implementation @see java.beans.VetoableChangeSupport#removeVetoableChangeListener(java.beans.VetoableChangeListener)
   * @param listener the listener to be removed.
   */
  public void removeVetoableChangeListener (VetoableChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getVetoableChangeSupport ().removeVetoableChangeListener (listener);
  }

  /**
   * Removes a VetoableChangeListener from the GraphRepository.
   * For an easy implementation @see java.beans.VetoableChangeSupport#removeVetoableChangeListener(java.lang.String, java.beans.VetoableChangeListener)
   * @param propertyName the property name.
   * @param listener the listener to be removed.
   */
  public void removeVetoableChangeListener (String propertyName, VetoableChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getVetoableChangeSupport ().removeVetoableChangeListener (propertyName, listener);
  }

  /**
   * Reports a vetoable property update to any registered listeners.
   * For an easy implementation @see java.beans.VetoableChangeSupport#fireVetoableChange(java.lang.String, java.lang.Object, java.lang.Object)
   * @param propertyName the property name.
   * @param oldValue the old value.
   * @param newValue the new value.
   * @throws PropertyVetoException A PropertyVetoException is thrown when a proposed change to a  property represents an unacceptable value.
   */
  public void fireVetoableChange (String propertyName, Object oldValue, Object newValue)
    throws PropertyVetoException {

    // 
    assert propertyName != null;

    // 
    this.getVetoableChangeSupport ().fireVetoableChange (propertyName, oldValue, newValue);
  }

  // //////////////////////////////////////////////////////////
  // 
  // PropertyChangeListener methods
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
   */
  public void propertyChange (PropertyChangeEvent evt) {
    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY, null, this);
  }

  // //////////////////////////////////////////////////////////
  // 
  // Element change
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see java.util.Map#clear()
   */
  public void clear () {
    super.clear ();
    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY, null, this);
  }

  /**
   * @see north.util.ValidatedMap#force(java.lang.Object, java.lang.Object)
   */
  protected Object force (Object key, Object value) {

    // 
    assert key != null;

    assert value != null;

    // 
    Object result = super.force (key, value);

    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY_ELEMENT, result, value);

    return result;
  }

  /**
   * @see java.util.Map#put(java.lang.Object, java.lang.Object)
   */
  public Object put (Object key, Object value) {

    // 
    assert key != null;

    assert value != null;

    // 
    Object result = super.put (key, value);

    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY_ELEMENT, result, value);

    return result;
  }

  /**
   * @see java.util.Map#putAll(java.util.Map)
   */
  public void putAll (java.util.Map map) {
    super.putAll (map);
    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY, null, this);
  }

  /**
   * @see java.util.Map#remove(java.lang.Object)
   */
  public Object remove (Object key) {

    // 
    assert key != null;

    // 
    Object result = super.remove (key);

    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY_ELEMENT, result, null);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.util.ValidatedMap#replace(java.lang.Object, java.lang.Object)
   */
  public Object replace (Object key, Object value) {

    // 
    assert key != null;

    assert value != null;

    // 
    Object result = super.replace (key, value);

    this.firePropertyChange (GraphRepository.GRAPH_REPOSITORY_ELEMENT, result, value);

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Status Report
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Tells if exists a graph registered with the given id.
   * @param id the registration id.
   * @return either true or false.
   */
  public boolean hasRegisteredGraph (String id) {

    // 
    assert id != null;

    // 
    return this.containsKey (id);
  }

  // //////////////////////////////////////////////////////////
  // 
  // Basic Operations
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Register the graph.
   * @param graph the registered graph.
   */
  public void registerGraph (IGraph graph) {

    // 
    assert graph != null;

    if (!graph.hasIdentifier ()) {
      throw new IllegalArgumentException ("Graph hasn't an identifier defined!");
    }

    // 
    String id = graph.getIdentifier ();
    IGraph oldGraph = (IGraph) this.force (id, graph);

    if (oldGraph != null) {
      oldGraph.removePropertyChangeListener (this);
    }

    graph.addPropertyChangeListener (this);

    // 
    assert this.hasRegisteredGraph (id);

    // 
  }

  /**
   * Unregister a graph.
   * @param graph the graph being unregistered
   */
  public void unregisterGraph (IGraph graph) {

    // 
    assert graph != null;

    if (!graph.hasIdentifier ()) {
      throw new IllegalArgumentException ("Graph hasn't an identifier defined!");
    }

    String id = graph.getIdentifier ();

    if (!this.hasRegisteredGraph (id)) {
      throw new IllegalArgumentException ("No such graph: '" + id + "'!");
    }

    // 
    graph.removePropertyChangeListener (this);
    this.remove (id);

    // 
    assert !this.hasRegisteredGraph (id);

    // 
  }

  /**
   * Retrieve a registered graph.
   * @param id the registration id.
   * @return the registered graph.
   */
  public IGraph retrieveGraph (String id) {

    // 
    assert id != null;

    if (!this.hasRegisteredGraph (id)) {
      throw new IllegalArgumentException ("No such graph: '" + id + "'!");
    }

    // 
    IGraph result = (IGraph) this.get (id);

    // 
    assert result != null;

    assert result.getIdentifier ().equals (id);

    // 
    return result;
  }

  /**
   * Generate an non registered identifier for a graph.
   * @param prefix the prefix to the generated identifier.
   * @return the generated identifier.
   */
  public String generateId (String prefix) {

    String result = null;

    for (int i = 1; result == null; i++) {
      String id = prefix + i;

      if (!this.hasRegisteredGraph (id)) {
        result = id;
      }
    }

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Output
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see java.lang.Object#toString()
   */
  public String toString () {

    StringBuffer buffer = new StringBuffer ("Registered graphs:\n");

    for (Iterator iter = this.keySet ().iterator (); iter.hasNext (); ) {
      buffer.append ("\t");
      buffer.append (iter.next ().toString ());
    }

    String result = buffer.toString ();

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** The singleton instance of the repository of graphs */
  private static final GraphRepository INSTANCE = GraphRepository.newGraphRepository ();

  /** The support object for the JavaBeans interface. */
  private PropertyChangeSupport _propertyChangeSupport = new PropertyChangeSupport (this);

  /** The vetoable support object for the JavaBeans interface. */
  private VetoableChangeSupport _vetoableChangeSupport = new VetoableChangeSupport (this);

  /** Event put */
  public static final String GRAPH_REPOSITORY = "GRAPH_REPOSITORY";

  /** Event put all */
  public static final String GRAPH_REPOSITORY_ELEMENT = "GRAPH_REPOSITORY_ELEMENT";
}


/*--- Formatted in North Style ---*/
