/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core.basic;



import north.graph.core.IArc;
import north.graph.core.IArcProvider;
import north.graph.core.IDOM;
import north.graph.core.IExportableArc;
import north.graph.core.IExportableNode;
import north.graph.core.INode;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * BasicArc class.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: BasicArc.java,v 1.15 2004/12/02 15:08:26 frederic Exp $
 */
public class BasicArc
  extends BasicCharacterized
  implements IExportableArc {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Constructor.
   */
  protected BasicArc () {
  }

  /**
   * Creates a new instance.
   * @param tail the tail node.
   * @param head the head node.
   * @return the new instance.
   */
  public static BasicArc newBasicArc (INode tail, INode head) {

    // 
    assert tail != null;

    assert head != null;

    // 
    BasicArc result = new BasicArc ();

    result.init (tail, head);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Creates a new instance.
   * @param tail the tail node.
   * @param head the head node.
   * @param identifier the arc identifier.
   * @param label the arc label.
   * @param type the arc's type.
   * @return the new instance.
   */
  public static BasicArc newBasicArc (INode tail, INode head, String identifier, String label, String type) {

    // 
    assert tail != null;

    assert head != null;

    assert identifier != null;

    assert label != null;

    assert type != null;

    // 
    BasicArc result = new BasicArc ();

    result.init (tail, head, identifier, label, type);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Init method.
   * @param tail the tail node.
   * @param head the head node.
   */
  protected void init (INode tail, INode head) {

    // 
    assert tail != null;

    assert head != null;

    assert tail.getGraph () == head.getGraph ();

    // 
    this.setTail (tail);
    this.setHead (head);

    // 
    assert this.getTail () == tail;

    assert this.getHead () == head;

    // 
  }

  /**
   * Init method.
   * @param tail the tail node.
   * @param head the head node.
   * @param identifier the identifier of the arc.
   * @param label the label of the arc.
   * @param type the arc's type.
   */
  protected void init (INode tail, INode head, String identifier, String label, String type) {

    // 
    assert tail != null;

    assert head != null;

    assert identifier != null;

    assert label != null;

    assert type != null;

    assert tail.getGraph () == head.getGraph ();

    // 
    this.init (tail, head);
    super.init (identifier, label, type);

    // 
    assert this.getTail () == tail;

    assert this.getHead () == head;

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.IArc#getHead()
   */
  public INode getHead () {

    // 
    assert this._head != null;

    // 
    return this._head;
  }

  /**
   * Sets the node at the head of the arc.
   * @param head the node at the head of the arc.
   */
  public void setHead (INode head) {

    // 
    assert head != null;

    // 
    this._head = head;

    // 
    assert this.getHead () == head;
  }

  /**
   * @see north.graph.core.IArc#getTail()
   */
  public INode getTail () {

    // 
    assert this._tail != null;

    // 
    return this._tail;
  }

  /**
   * Sets the node at the tail of the arc.
   * @param tail the node at the tail of the arc.
   */
  public void setTail (INode tail) {

    // 
    assert tail != null;

    // 
    this._tail = tail;

    // 
    assert this.getTail () == tail;

    // 
  }

  /**
   * @see north.graph.core.basic.BasicCharacterized#getDOMTagName()
   */
  protected String getDOMTagName () {
    return IDOM.ARC;
  }

  /**
   * @see north.graph.core.IArcProvider#getArc()
   */
  public IArc getArc () {
    return this;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Comparison
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Tells if the technical identifier is the same in both objects.
   * @param object the object to be tested.
   * @return true if the technical identifier is the same; otherwise, false.
   */
  public boolean equals (Object object) {

    boolean result = false;

    if (object instanceof IArcProvider) {
      IArc arc = ((IArcProvider) object).getArc ();

      result = this == arc;
    }

    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Output
  // 
  // ////////////////////////////////////////////////////////  //

  /**
   * @see north.graph.core.IExportable#toElement(org.w3c.dom.Document)
   */
  public Element toElement (Document document) {

    // 
    assert document != null;

    // 
    // 
    // Create the DOM element
    // 
    Element result = super.toElement (document);
    String tailId = ((IExportableNode) this.getTail ()).getIdentifier ();
    String headId = ((IExportableNode) this.getHead ()).getIdentifier ();

    result.setAttribute (IDOM.ID, tailId + "-" + headId);

    if (this.hasType ()) {
      result.setAttribute (IDOM.TYPE, this.getType ());
    }

    result.setAttribute (IDOM.TAIL, tailId);
    result.setAttribute (IDOM.HEAD, headId);

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** the head node of the arc */
  private INode _head = null;

  /** the tail node of the arc */
  private INode _tail = null;
}


/*--- Formatted in North Style ---*/
