/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core.basic;



import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Map.Entry;

import north.graph.core.CharacterizationSupport;
import north.graph.core.ICharacterized;
import north.graph.core.IDOM;
import north.graph.core.IData;
import north.graph.core.IEditableData;
import north.graph.core.IExportable;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * BasicCharacterized abstract class which is the basic implementation of the ICharacterized.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: BasicCharacterized.java,v 1.15 2004/10/13 07:35:10 fabian Exp $
 */
public abstract class BasicCharacterized
  implements ICharacterized, IExportable {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Init method.
   */
  protected void init () {
  }

  /**
   * Init method.
   * @param identifier the identifier of the object.
   * @param label the label of the object.
   * @param type the type of the object.
   */
  protected void init (String identifier, String label, String type) {

    // 
    assert identifier != null;

    assert label != null;

    assert type != null;

    // 
    this.setIdentifier (identifier);
    this.setLabel (label);
    this.setType (type);

    // 
    assert this.getIdentifier () == identifier;

    assert this.getLabel () == label;

    assert this.getType () == type;

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.ICharacterized#getIdentifier()
   */
  public String getIdentifier () {

    if (!this.hasIdentifier ()) {
      throw new IllegalStateException ("The identifier has not been set!");
    }

    String result = (String) this.getValue (ICharacterized.IDENTIFIER_KEY);

    return result;
  }

  /**
   * Sets the identifier.
   * @param id the identifier value.
   */
  protected void setIdentifier (String id) {

    // 
    assert id != null;

    // 
    this.putAttribute (ICharacterized.IDENTIFIER_KEY, id);

    // 
    assert this.getIdentifier () == id;

    // 
  }

  /**
   * @see north.graph.core.ICharacterized#getLabel()
   */
  public String getLabel () {

    if (!this.hasLabel ()) {
      throw new IllegalStateException ("The label has not been set!");
    }

    String result = (String) this.getValue (ICharacterized.LABEL_KEY);

    return result;
  }

  /**
   * Sets the label.
   * @param label the identifier value.
   */
  public void setLabel (String label) {

    // 
    assert label != null;

    // 
    this.putAttribute (ICharacterized.LABEL_KEY, label);

    // 
    assert this.getLabel () == label;

    // 
  }

  /**
   * @see north.graph.core.ICharacterized#getType()
   */
  public String getType () {

    if (!this.hasType ()) {
      throw new IllegalStateException ("The type has not been set!");
    }

    return this._type;
  }

  /**
   * Sets the type of the characterized object.
   * @param type the type to be set.
   */
  protected void setType (String type) {

    // 
    assert type != null;

    // 
    this._type = type;

    // 
    assert this.getType () == type;
  }

  /**
   * Returns the attributes map.
   * @return the attributes map.
   */
  public HashMap getAttributeMap () {

    // 
    assert this._attributeMap != null;

    // 
    return this._attributeMap;
  }

  /**
   * The attributes to set.
   * @param attributeMap the given attributes.
   */
  protected void setAttributeMap (HashMap attributeMap) {

    // 
    assert attributeMap != null;

    // 
    this._attributeMap = attributeMap;

    // 
    assert this.getAttributeMap () == attributeMap;

    // 
  }

  /**
   * @see north.graph.core.ICharacterized#getValue(java.lang.String)
   */
  public Object getValue (String key) {

    // 
    assert key != null;

    assert this.containsAttributeKey (key);

    // 
    Object result = this.getAttributeMap ().get (key);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.ICharacterized#getValue(north.graph.core.IData, java.lang.String)
   */
  public Object getValue (IData data, String key) {

    // 
    assert data != null;

    assert key != null;

    assert this.containsAttribute (data, key);

    // 
    Object result = CharacterizationSupport.getValue (this, data, key);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.ICharacterized#getKeys()
   */
  public Set getKeys () {

    Set result = new HashSet (this.getAttributeMap ().keySet ());

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Returns the XML Document Object Model tag name.
   * @return the tag name.
   */
  protected abstract String getDOMTagName ();

  // //////////////////////////////////////////////////////////
  // 
  // Status
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.ICharacterized#hasIdentifier()
   */
  public boolean hasIdentifier () {
    return this.containsAttribute (ICharacterized.IDENTIFIER_KEY);
  }

  /**
   * @see north.graph.core.ICharacterized#hasLabel()
   */
  public boolean hasLabel () {
    return this.containsAttribute (ICharacterized.LABEL_KEY);
  }

  /**
   * @see north.graph.core.ICharacterized#hasType()
   */
  public boolean hasType () {
    return this._type != null;
  }

  /**
   * @see north.graph.core.ICharacterized#containsAttribute(java.lang.String)
   */
  public boolean containsAttribute (String key) {

    // 
    assert key != null;

    // 
    boolean result = this.containsAttributeKey (key);

    return result;
  }

  /**
   * @see north.graph.core.ICharacterized#containsAttribute(north.graph.core.IData, java.lang.String)
   */
  public boolean containsAttribute (IData data, String key) {

    // 
    assert data != null;

    assert key != null;

    // 
    boolean result = CharacterizationSupport.containsAttribute (this, data, key);

    return result;
  }

  /**
   * Tells if an attribute associated to the given key is present in the current characterized object.
   * @param key the key of the attribute.
   * @return true if an attribute associated to the given key is present in the current characterized object.
   */
  public boolean containsAttributeKey (String key) {

    // 
    assert key != null;

    // 
    boolean result = this.getAttributeMap ().containsKey (key);

    return result;
  }

  /**
   * Tells if an attribute with the given value is present in the current characterized object.
   * @param value the given value.
   * @return true if an attribute with the given value is present in the current characterized object.
   */
  protected boolean containsAttributeValue (Object value) {

    // 
    assert value != null;

    // 
    boolean result = this.getAttributeMap ().containsValue (value);

    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Element change
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Adds an attribute to the characterized object.
   * @param key the key to be used.
   * @param value the value to be inserted.
   */
  public void putAttribute (String key, Object value) {

    // 
    assert key != null;

    assert value != null;

    // 
    this.getAttributeMap ().put (key, value);
  }

  /**
   * @see north.graph.core.ICharacterized#putAttribute(north.graph.core.IEditableData, java.lang.String, java.lang.Object)
   */
  public void putAttribute (IEditableData data, String key, Object value) {

    // 
    assert data != null;

    assert key != null;

    assert value != null;

    // 
    CharacterizationSupport.putAttribute (this, data, key, value);
  }

  /**
   * Removes an attribute from the characterized object.
   * @param key the key of the attribute to be removed.
   */
  public void removeAttribute (String key) {

    // 
    assert key != null;

    assert this.containsAttributeKey (key);

    // 
    this.getAttributeMap ().remove (key);

    // 
    assert !this.containsAttributeKey (key);

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Output
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.IExportable#toElement(org.w3c.dom.Document)
   */
  public Element toElement (Document document) {

    // 
    assert document != null;

    // 
    // 
    // Create the DOM element
    // 
    Element result = document.createElement (this.getDOMTagName ());

    // 
    // Attributes
    // 
    for (Iterator iter = this.getAttributeMap ().entrySet ().iterator (); iter.hasNext (); ) {
      Entry entry = (Entry) iter.next ();
      String attribute = (String) entry.getKey ();
      Object value = entry.getValue ();

      if (attribute != ICharacterized.IDENTIFIER_KEY) {
        Element attrElt = document.createElement (IDOM.ATTR);

        result.appendChild (attrElt);
        attrElt.setAttribute (IDOM.NAME, attribute);
        attrElt.setAttribute (IDOM.TYPE, value.getClass ().getName ());

        if (value != null) {
          attrElt.setAttribute (IDOM.VALUE, value.toString ());
        }
      }
    }

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** the type of the characterized object */
  private String _type = null;

  /** the attributes map of the characterized object */
  private HashMap _attributeMap = new HashMap ();
}


/*--- Formatted in North Style ---*/
