/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core.basic;



import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import north.graph.core.ICharacterized;
import north.graph.core.IDOM;
import north.graph.core.IExportableArc;
import north.graph.core.IExportableGraph;
import north.graph.core.IExportableNode;
import north.graph.core.IExportableSubdividedGraph;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * BasicGraph class.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: BasicGraph.java,v 1.25 2004/11/16 15:49:23 frederic Exp $
 */
public class BasicGraph
  extends BasicCharacterized
  implements IExportableSubdividedGraph {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Constructor.
   */
  protected BasicGraph () {
  }

  /**
   * Creates a new instance.
   * @return the new instance.
   */
  public static BasicGraph newBasicGraph () {

    BasicGraph result = new BasicGraph ();

    result.init ();

    return result;
  }

  /**
   * Creates a new instance.
   * @param identifier the identifier of the graph.
   * @param label the label of the graph.
   * @param type the graph type.
   * @return the new instance.
   */
  public static BasicGraph newBasicGraph (String identifier, String label, String type) {

    // 
    assert identifier != null;

    assert label != null;

    assert type != null;

    // 
    BasicGraph result = new BasicGraph ();

    result.init (identifier, label, type);

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Returns the nodes map.
   * @return the nodes map.
   */
  protected HashMap getNodeMap () {

    // 
    assert this._nodeMap != null;

    // 
    return this._nodeMap;
  }

  /**
   * Sets nodes map.
   * @param nodeMap the nodes map.
   */
  protected void setNodeMap (HashMap nodeMap) {

    // 
    assert nodeMap != null;

    // 
    this._nodeMap = nodeMap;

    // 
    assert this.getNodeMap () == nodeMap;

    // 
  }

  /**
   * Returns the arcs set.
   * @return the arcs set.
   */
  protected HashSet getArcSet () {

    // 
    assert this._arcSet != null;

    // 
    return this._arcSet;
  }

  /**
   * Sets the arcs set.
   * @param arcSet the arcs set.
   */
  protected void setArcSet (HashSet arcSet) {

    // 
    assert arcSet != null;

    // 
    this._arcSet = arcSet;

    // 
    assert this.getArcSet () == arcSet;

    // 
  }

  /**
   * Returns the node corresponding to the given identifier.
   * @param nodeId a node identifier.
   * @return the corresponding node.
   */
  public BasicNode getNode (String nodeId) {

    // 
    assert nodeId != null;

    assert this.hasNode (nodeId);

    // 
    BasicNode result = (BasicNode) this.getNodeMap ().get (nodeId);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Returns the incoming arcs set for the given node.
   * @param node the given node.
   * @return the arcs set.
   */
  public Set getInArcs (BasicNode node) {

    // 
    assert node != null;

    assert this.hasNode (node);

    // 
    Set result = new HashSet ();

    for (Iterator iter = this.getArcSet ().iterator (); iter.hasNext (); ) {
      BasicArc arc = (BasicArc) iter.next ();

      if (arc.getHead () == node) {
        result.add (arc);
      }
    }

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Returns the outgoing arcs set for the given node.
   * @param node the given node.
   * @return the arcs set.
   */
  public Set getOutArcs (BasicNode node) {

    // 
    assert node != null;

    assert this.hasNode (node);

    // 
    Set result = new HashSet ();

    for (Iterator iter = this.getArcSet ().iterator (); iter.hasNext (); ) {
      BasicArc arc = (BasicArc) iter.next ();

      if (arc.getTail () == node) {
        result.add (arc);
      }
    }

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Returns the subgraphList.
   * @return the subgraphList.
   */
  protected LinkedList getSubgraphList () {

    // 
    assert this._subgraphList != null;

    // 
    return this._subgraphList;
  }

  /**
   * The subgraphList to set.
   * @param subgraphList the given subgraphList.
   */
  protected void setSubgraphList (LinkedList subgraphList) {

    // 
    assert subgraphList != null;

    // 
    this._subgraphList = subgraphList;

    // 
    assert this.getSubgraphList () == subgraphList;

    // 
  }

  /**
   * @see north.graph.core.IGraph#getNodes()
   */
  public Set getNodes () {
    return Collections.unmodifiableSet (new HashSet (this.getNodeMap ().values ()));
  }

  /**
   * @see north.graph.core.IGraph#getArcs()
   */
  public Set getArcs () {
    return Collections.unmodifiableSet (this.getArcSet ());
  }

  /**
   * @see north.graph.core.ISubdividedGraph#getSubgraphs()
   */
  public List getSubgraphs () {
    return Collections.unmodifiableList (this.getSubgraphList ());
  }

  /**
   * @see north.graph.core.ICharacterized#getKeys()
   */
  public Set getKeys () {

    Set result = super.getKeys ();

    // 
    // return the set of keys without the "identifier" key.
    // 
    if (result.contains (ICharacterized.IDENTIFIER_KEY)) {
      result.remove (ICharacterized.IDENTIFIER_KEY);
    }

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.basic.BasicCharacterized#getDOMTagName()
   */
  protected String getDOMTagName () {
    return IDOM.GRAPH;
  }

  /**
   * Returns the propertyChangeSupport.
   * @return the propertyChangeSupport.
   */
  protected PropertyChangeSupport getPropertyChangeSupport () {

    // 
    assert this._propertyChangeSupport != null;

    // 
    return this._propertyChangeSupport;
  }

  /**
   * Returns the vetoableChangeSupport.
   * @return the vetoableChangeSupport.
   */
  protected VetoableChangeSupport getVetoableChangeSupport () {

    // 
    assert this._vetoableChangeSupport != null;

    // 
    return this._vetoableChangeSupport;
  }

  // //////////////////////////////////////////////////////////
  // 
  // JavaBeans
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.IGraph#addPropertyChangeListener(java.beans.PropertyChangeListener)
   */
  public void addPropertyChangeListener (PropertyChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getPropertyChangeSupport ().addPropertyChangeListener (listener);
  }

  /**
   * @see north.graph.core.IGraph#addPropertyChangeListener(java.lang.String, java.beans.PropertyChangeListener)
   */
  public void addPropertyChangeListener (String propertyName, PropertyChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getPropertyChangeSupport ().addPropertyChangeListener (propertyName, listener);
  }

  /**
   * @see north.graph.core.IGraph#removePropertyChangeListener(java.beans.PropertyChangeListener)
   */
  public void removePropertyChangeListener (PropertyChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getPropertyChangeSupport ().removePropertyChangeListener (listener);
  }

  /**
   * @see north.graph.core.IGraph#removePropertyChangeListener(java.lang.String, java.beans.PropertyChangeListener)
   */
  public void removePropertyChangeListener (String propertyName, PropertyChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getPropertyChangeSupport ().removePropertyChangeListener (propertyName, listener);
  }

  /**
   * @see north.graph.core.IGraph#firePropertyChange(java.lang.String, java.lang.Object, java.lang.Object)
   */
  public void firePropertyChange (String propertyName, Object oldValue, Object newValue) {

    // 
    assert propertyName != null;

    // 
    this.getPropertyChangeSupport ().firePropertyChange (propertyName, oldValue, newValue);
  }

  /**
   * @see north.graph.core.IGraph#addVetoableChangeListener(java.beans.VetoableChangeListener)
   */
  public void addVetoableChangeListener (VetoableChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getVetoableChangeSupport ().addVetoableChangeListener (listener);
  }

  /**
   * @see north.graph.core.IGraph#addVetoableChangeListener(java.lang.String, java.beans.VetoableChangeListener)
   */
  public void addVetoableChangeListener (String propertyName, VetoableChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getVetoableChangeSupport ().addVetoableChangeListener (propertyName, listener);
  }

  /**
   * @see north.graph.core.IGraph#removeVetoableChangeListener(java.beans.VetoableChangeListener)
   */
  public void removeVetoableChangeListener (VetoableChangeListener listener) {

    // 
    assert listener != null;

    // 
    this.getVetoableChangeSupport ().removeVetoableChangeListener (listener);
  }

  /**
   * @see north.graph.core.IGraph#removeVetoableChangeListener(java.lang.String, java.beans.VetoableChangeListener)
   */
  public void removeVetoableChangeListener (String propertyName, VetoableChangeListener listener) {

    // 
    assert propertyName != null;

    assert listener != null;

    // 
    this.getVetoableChangeSupport ().removeVetoableChangeListener (propertyName, listener);
  }

  /**
   * @throws PropertyVetoException
   * @see north.graph.core.IGraph#fireVetoableChange(java.lang.String, java.lang.Object, java.lang.Object)
   */
  public void fireVetoableChange (String propertyName, Object oldValue, Object newValue)
    throws PropertyVetoException {

    // 
    assert propertyName != null;

    // 
    this.getVetoableChangeSupport ().fireVetoableChange (propertyName, oldValue, newValue);
  }

  // //////////////////////////////////////////////////////////
  // 
  // Status report
  // 
  // ////////////////////////////////////////////////////////  //

  /**
   * Tells if the given node is present in this graph.
   * @param node the tested node.
   * @return true if the given node is present in the current graph.
   */
  public boolean hasNode (BasicNode node) {

    // 
    assert node != null;

    // 
    boolean result = this.getNodeMap ().containsValue (node);

    return result;
  }

  /**
   * Tells if the given node identifier is present in this graph.
   * @param nodeId the tested identifier.
   * @return true if the given identifier is present in this graph.
   */
  public boolean hasNode (String nodeId) {

    // 
    assert nodeId != null;

    // 
    boolean result = this.getNodeMap ().containsKey (nodeId);

    return result;
  }

  /**
   * Tells if the given arc is present in this graph.
   * @param arc the tested arc.
   * @return true if the given arc is present in this graph.
   */
  public boolean hasArc (BasicArc arc) {

    // 
    assert arc != null;

    // 
    boolean result = this.getArcSet ().contains (arc);

    return result;
  }

  /**
   * Tells if the given subgraph is present in this graph.
   * @param subgraph the tested subgraph.
   * @return true if the given subgraph is present in this graph.
   */
  public boolean hasSubgraph (BasicSubgraph subgraph) {

    // 
    assert subgraph != null;

    // 
    boolean result = this.getSubgraphList ().contains (subgraph);

    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Element change
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Adds a node to this graph.
   * @param node the node to be added.
   */
  public void addNode (BasicNode node) {

    // 
    assert node != null;

    assert node.getGraph () == this;

    assert !this.hasNode (node.getIdentifier ());

    // 
    this.getNodeMap ().put (node.getIdentifier (), node);

    // 
    assert this.hasNode (node.getIdentifier ());

    // 
  }

  /**
   * Adds an arc to this graph.
   * @param arc the arc to be added.
   */
  public void addArc (BasicArc arc) {

    // 
    assert arc != null;

    assert !this.hasArc (arc);

    // 
    this.getArcSet ().add (arc);

    // 
    assert this.hasArc (arc);

    // 
  }

  /**
   * Adds an subgraph to this graph.
   * @param subgraph the Subgraph to be added.
   */
  public void addSubgraph (BasicSubgraph subgraph) {

    // 
    assert subgraph != null;

    assert !this.hasSubgraph (subgraph);

    // 
    this.getSubgraphList ().add (subgraph);

    // 
    assert this.hasSubgraph (subgraph);

    assert this.hasSubgraph (subgraph);

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Output
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.IExportable#toElement(org.w3c.dom.Document)
   */
  public Element toElement (Document document) {

    // 
    assert document != null;

    // 
    // 
    // Create the DOM element
    // 
    Element result = super.toElement (document);
    String id = this.hasIdentifier ()
                ? this.getIdentifier ()
                : "default";

    result.setAttribute (IDOM.ID, id);
    result.setAttribute (IDOM.TYPE, this.getClass ().getName ());

    // Nodes
    for (Iterator iter = this.getNodeMap ().values ().iterator (); iter.hasNext (); ) {
      IExportableNode node = (IExportableNode) iter.next ();

      result.appendChild (node.toElement (document));
    }

    // Arcs
    for (Iterator iter = this.getArcSet ().iterator (); iter.hasNext (); ) {
      IExportableArc arc = (IExportableArc) iter.next ();

      result.appendChild (arc.toElement (document));
    }

    // Subgraphs
    for (Iterator iter = this.getSubgraphList ().iterator (); iter.hasNext (); ) {
      IExportableGraph subgraph = (IExportableGraph) iter.next ();

      result.appendChild (subgraph.toElement (document));
    }

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** The nodes map */
  private HashMap _nodeMap = new HashMap ();

  /** The arcs set */
  private HashSet _arcSet = new HashSet ();

  /** The list of subgraphs */
  private LinkedList _subgraphList = new LinkedList ();

  /** The support object for the JavaBeans interface. */
  private PropertyChangeSupport _propertyChangeSupport = new PropertyChangeSupport (this);

  /** The vetoable support object for the JavaBeans interface. */
  private VetoableChangeSupport _vetoableChangeSupport = new VetoableChangeSupport (this);
}


/*--- Formatted in North Style ---*/
