/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core.basic;



import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import north.graph.core.ICharacterized;
import north.graph.core.IDOM;
import north.graph.core.IExportableNode;
import north.graph.core.IGraph;
import north.graph.core.INode;
import north.graph.core.INodeProvider;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * BasicNode class.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: BasicNode.java,v 1.20 2004/12/02 15:08:26 frederic Exp $
 */
public class BasicNode
  extends BasicCharacterized
  implements IExportableNode {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Constructor.
   */
  protected BasicNode () {
  }

  /**
   * Creates a new instance.
   * @param graph the graph.
   * @param identifier the identifier of the node.
   * @param label the label of the node.
   * @param type the type of the node.
   * @return the new instance.
   */
  public static BasicNode newBasicNode (IGraph graph, String identifier, String label, String type) {

    // 
    assert graph != null;

    assert identifier != null;

    assert label != null;

    assert type != null;

    // 
    BasicNode result = new BasicNode ();

    result.init (graph, identifier, label, type);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * Init method.
   * @param graph the graph.
   * @param identifier the identifier of the object.
   * @param label the label of the object.
   * @param type the type of the object.
   */
  protected void init (IGraph graph, String identifier, String label, String type) {

    // 
    assert graph != null;

    assert identifier != null;

    assert label != null;

    assert type != null;

    // 
    this.setGraph (graph);
    super.init (identifier, label, type);

    // 
    assert this.getGraph () == graph;

    assert this.getIdentifier () == identifier;

    assert this.getLabel () == label;

    assert this.getType () == type;

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.INode#getGraph()
   */
  public IGraph getGraph () {

    // 
    assert this._graph != null;

    // 
    return this._graph;
  }

  /**
   * The graph to set.
   * @param graph the given graph.
   */
  protected void setGraph (IGraph graph) {

    // 
    assert graph != null;

    // 
    this._graph = graph;

    // 
    assert this.getGraph () == graph;

    // 
  }

  /**
   * @see north.graph.core.ICharacterized#getKeys()
   */
  public Set getKeys () {

    Set result = super.getKeys ();

    // 
    // return the set of keys without the "identifier" key.
    // 
    if (result.contains (ICharacterized.IDENTIFIER_KEY)) {
      result.remove (ICharacterized.IDENTIFIER_KEY);
    }

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.basic.BasicCharacterized#getDOMTagName()
   */
  protected String getDOMTagName () {
    return IDOM.NODE;
  }

  /**
   * @see north.graph.core.INode#getInArcs()
   */
  public Set getInArcs () {

    Set result = null;
    Set arcs = this.getGraph ().getArcs ();
    HashSet inArcs = new HashSet ();

    for (Iterator iter = arcs.iterator (); iter.hasNext (); ) {
      BasicArc arc = (BasicArc) iter.next ();

      if (arc.getHead () == this) {
        inArcs.add (arc);
      }
    }

    result = Collections.unmodifiableSet (inArcs);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.INode#getOutArcs()
   */
  public Set getOutArcs () {

    Set result = null;
    Set arcs = this.getGraph ().getArcs ();
    HashSet outArcs = new HashSet ();

    for (Iterator iter = arcs.iterator (); iter.hasNext (); ) {
      BasicArc arc = (BasicArc) iter.next ();

      if (arc.getTail () == this) {
        outArcs.add (arc);
      }
    }

    result = Collections.unmodifiableSet (outArcs);

    // 
    assert result != null;

    // 
    return result;
  }

  /**
   * @see north.graph.core.INodeProvider#getNode()
   */
  public INode getNode () {
    return this;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Comparison
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Tells if the technical identifier is the same in both objects.
   * @param object the object to be tested.
   * @return true if the technical identifier is the same; otherwise, false.
   */
  public boolean equals (Object object) {

    boolean result = false;

    if (object instanceof INodeProvider) {
      INode node = ((INodeProvider) object).getNode ();

      result = this == node;
    }

    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Output
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.IExportable#toElement(org.w3c.dom.Document)
   */
  public Element toElement (Document document) {

    // 
    assert document != null;

    // 
    // 
    // Create the DOM element
    // 
    Element result = super.toElement (document);

    result.setAttribute (IDOM.ID, this.getIdentifier ());

    if (this.hasType ()) {
      result.setAttribute (IDOM.TYPE, this.getType ());
    }

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** The graph */
  private IGraph _graph = null;
}


/*--- Formatted in North Style ---*/
