/*******************************************************************************
* Copyright (c) 2002-2004, North Bears-SCMBB-ULB
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*    - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*    - Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* Neither the names of North Bears, SCMBB or ULB nor the names of their
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
 *******************************************************************************/

package north.graph.core.builder;



import java.util.Calendar;

import north.graph.core.IGraph;
import north.graph.core.basic.BasicArc;
import north.graph.core.basic.BasicGraph;
import north.graph.core.basic.BasicNode;
import north.graph.core.basic.BasicSubgraph;


/**
 * MetabolicGraphBuilder class.
 *
 * @author Erick Antezana, Fabian Couche, Didier Croes, Simon De Keyzer, Frederic Fays, Olivier Hubaut and Christian Lemer.
 * @version $Id: MetabolicGraphBuilder.java,v 1.10 2004/11/16 15:51:16 frederic Exp $
 */
public class MetabolicGraphBuilder
  implements IMetabolicGraphBuilder {

  // //////////////////////////////////////////////////////////
  // 
  // Initialization
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Constructor.
   */
  protected MetabolicGraphBuilder () {
  }

  /**
   * Creates a new instance.
   * @return the new instance.
   */
  public static MetabolicGraphBuilder newMetabolicGraphBuilder () {

    MetabolicGraphBuilder result = new MetabolicGraphBuilder ();

    result.init ();

    return result;
  }

  /**
   * Init method.
   */
  protected void init () {

    BasicGraph graph = BasicGraph.newBasicGraph ();

    this.setBasicGraph (graph);

    // 
    assert this.getBasicGraph () == graph;

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Access
  // 
  // //////////////////////////////////////////////////////////

  /**
   * Returns the graph.
   * @return the graph.
   */
  protected BasicGraph getBasicGraph () {

    // 
    assert this._basicGraph != null;

    // 
    return this._basicGraph;
  }

  /**
   * The graph to set.
   * @param graph the given graph.
   */
  protected void setBasicGraph (BasicGraph graph) {

    // 
    assert graph != null;

    // 
    this._basicGraph = graph;

    // 
    assert this.getBasicGraph () == graph;

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#getGraph()
   */
  public IGraph getGraph () {

    IGraph result = this.getBasicGraph ();

    // 
    assert result != null;

    // 
    return result;
  }

  // //////////////////////////////////////////////////////////
  // 
  // Basic operations
  // 
  // //////////////////////////////////////////////////////////

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildGene(java.lang.String)
   */
  public void buildGene (String geneId) {

    // 
    assert geneId != null;

    assert !this.getBasicGraph ().hasNode (geneId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicNode node = BasicNode.newBasicNode (graph, geneId, geneId, GENE_TYPE);

    node.putAttribute ("fillcolor", "red");
    graph.addNode (node);

    // 
    assert this.getBasicGraph ().hasNode (geneId);

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildPolypeptide(java.lang.String)
   */
  public void buildPolypeptide (String polypeptideId) {

    // 
    assert polypeptideId != null;

    assert !this.getBasicGraph ().hasNode (polypeptideId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicNode node = BasicNode.newBasicNode (graph, polypeptideId, polypeptideId, POLYPEPTIDE_TYPE);

    node.putAttribute ("fillcolor", "blue");
    node.putAttribute ("shape", "box");
    graph.addNode (node);

    // 
    assert this.getBasicGraph ().hasNode (polypeptideId);

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildCompound(java.lang.String)
   */
  public void buildCompound (String compoundId) {

    // 
    assert compoundId != null;

    assert !this.getBasicGraph ().hasNode (compoundId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicNode node = BasicNode.newBasicNode (graph, compoundId, compoundId, COMPOUND_TYPE);

    node.putAttribute ("fillcolor", "purple");
    graph.addNode (node);

    // 
    assert this.getBasicGraph ().hasNode (compoundId);

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildExpression(java.lang.String, java.lang.String, java.lang.String)
   */
  public void buildExpression (String expressionId, String geneId, String polypeptideId) {

    // 
    assert expressionId != null;

    assert geneId != null;

    assert polypeptideId != null;

    assert !this.getBasicGraph ().hasNode (expressionId);

    assert this.getBasicGraph ().hasNode (geneId);

    assert this.getBasicGraph ().hasNode (polypeptideId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicNode geneNode = graph.getNode (geneId);

    assert geneNode.getType () == GENE_TYPE;

    BasicNode polypeptideNode = graph.getNode (polypeptideId);

    assert polypeptideNode.getType () == POLYPEPTIDE_TYPE;

    BasicNode expressionNode = BasicNode.newBasicNode (graph, expressionId, expressionId, EXPRESSION_TYPE);

    expressionNode.putAttribute ("shape", "box");
    graph.addNode (expressionNode);

    BasicArc geneExpressionArc = BasicArc.newBasicArc (geneNode, expressionNode);

    graph.addArc (geneExpressionArc);

    BasicArc expressionPolypeptideArc = BasicArc.newBasicArc (expressionNode, polypeptideNode);

    graph.addArc (expressionPolypeptideArc);

    BasicSubgraph subgraph = BasicSubgraph.newBasicSubgraph (graph, expressionId + geneId + polypeptideId, "",
                                                             EXPRESSION_TYPE);

    subgraph.addNode (geneNode);
    subgraph.addNode (polypeptideNode);
    subgraph.addNode (expressionNode);
    subgraph.addArc (geneExpressionArc);
    subgraph.addArc (expressionPolypeptideArc);

    // subgraph.putAttribute ("rank", "same");
    graph.addSubgraph (subgraph);

    // 
    assert graph.hasNode (expressionId);

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildReaction(java.lang.String, java.lang.String[], java.lang.String[])
   */
  public void buildReaction (String reactionId, String [] substrateIds, String [] productIds) {

    // 
    assert reactionId != null;

    assert substrateIds != null;

    assert productIds != null;

    assert !this.getBasicGraph ().hasNode (reactionId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicSubgraph reactionSubgraph =
      BasicSubgraph.newBasicSubgraph (graph, new Long (Calendar.getInstance ().getTimeInMillis ()).toString (), "",
                                      REACTION_TYPE);
    BasicSubgraph substrateSubgraph =
      BasicSubgraph.newBasicSubgraph (graph, new Long (Calendar.getInstance ().getTimeInMillis ()).toString (), "",
                                      SUBSTRATE_TYPE);
    BasicSubgraph productSubgraph =
      BasicSubgraph.newBasicSubgraph (graph, new Long (Calendar.getInstance ().getTimeInMillis ()).toString (), "",
                                      PRODUCT_TYPE);
    BasicNode reactionNode = BasicNode.newBasicNode (graph, reactionId, reactionId, REACTION_TYPE);

    reactionNode.putAttribute ("shape", "box");
    graph.addNode (reactionNode);
    reactionSubgraph.addNode (reactionNode);

    // Add substrates
    for (int i = 0; i < substrateIds.length; i++) {
      String substrateId = substrateIds [i];

      assert graph.hasNode (substrateId);

      BasicNode substrateNode = graph.getNode (substrateId);

      reactionSubgraph.addNode (substrateNode);
      substrateSubgraph.addNode (substrateNode);

      assert substrateNode.getType () == COMPOUND_TYPE;

      BasicArc substrateReactionArc = BasicArc.newBasicArc (substrateNode, reactionNode);

      graph.addArc (substrateReactionArc);
      reactionSubgraph.addArc (substrateReactionArc);
    }

    // Add products
    for (int j = 0; j < productIds.length; j++) {
      String productId = productIds [j];

      assert graph.hasNode (productId);

      BasicNode productNode = graph.getNode (productId);

      reactionSubgraph.addNode (productNode);
      productSubgraph.addNode (productNode);

      assert productNode.getType () == COMPOUND_TYPE;

      BasicArc reactionProductArc = BasicArc.newBasicArc (reactionNode, productNode);

      graph.addArc (reactionProductArc);
      reactionSubgraph.addArc (reactionProductArc);
    }

    // substrateSubgraph.putAttribute ("rank", "same");
    // productSubgraph.putAttribute ("rank", "same");
    graph.addSubgraph (reactionSubgraph);
    graph.addSubgraph (substrateSubgraph);
    graph.addSubgraph (productSubgraph);

    // 
    assert graph.hasNode (reactionId);

    // 
  }

  /**
   * @see north.graph.core.builder.IMetabolicGraphBuilder#buildCatalysis(java.lang.String, java.lang.String, java.lang.String)
   */
  public void buildCatalysis (String catalysisId, String polypeptideId, String reactionId) {

    // 
    assert catalysisId != null;

    assert polypeptideId != null;

    assert reactionId != null;

    assert !this.getBasicGraph ().hasNode (catalysisId);

    assert this.getBasicGraph ().hasNode (polypeptideId);

    assert this.getBasicGraph ().hasNode (reactionId);

    // 
    BasicGraph graph = this.getBasicGraph ();
    BasicNode polypeptideNode = graph.getNode (polypeptideId);

    assert polypeptideNode.getType () == POLYPEPTIDE_TYPE;

    BasicNode reactionNode = graph.getNode (reactionId);

    assert reactionNode.getType () == REACTION_TYPE;

    BasicNode catalysisNode = BasicNode.newBasicNode (graph, catalysisId, catalysisId, CATALYSIS_TYPE);

    catalysisNode.putAttribute ("shape", "box");
    graph.addNode (catalysisNode);

    BasicArc polypeptideCatalysisArc = BasicArc.newBasicArc (polypeptideNode, catalysisNode);

    graph.addArc (polypeptideCatalysisArc);

    BasicArc catalysisReactionArc = BasicArc.newBasicArc (catalysisNode, reactionNode);

    graph.addArc (catalysisReactionArc);

    BasicSubgraph subgraph =
      BasicSubgraph.newBasicSubgraph (graph, new Long (Calendar.getInstance ().getTimeInMillis ()).toString (), "",
                                      CATALYSIS_TYPE);

    subgraph.addNode (polypeptideNode);
    subgraph.addNode (reactionNode);
    subgraph.addNode (catalysisNode);
    subgraph.addArc (polypeptideCatalysisArc);
    subgraph.addArc (catalysisReactionArc);

    // subgraph.putAttribute ("rank", "same");
    graph.addSubgraph (subgraph);

    // 
    assert graph.hasNode (catalysisId);

    // 
  }

  // //////////////////////////////////////////////////////////
  // 
  // Implementation
  // 
  // //////////////////////////////////////////////////////////

  /** The associated graph */
  private BasicGraph _basicGraph;
}


/*--- Formatted in North Style ---*/
